
import os
import socket
import sys
import time
from typing import TYPE_CHECKING
from PyQt5.QtCore import QObject, QProcess, QTimer
from PyQt5.QtWidgets import QApplication
try:
    from liblo import Address
except ImportError:
    from pyliblo3 import Address

import ray
from gui_server_thread import GuiServerThread
from gui_tools import CommandLineArgs, ErrDaemon, _translate

if TYPE_CHECKING:
    from .gui_session import SignaledSession


class DaemonManager(QObject):
    def __init__(self, session: 'SignaledSession'):
        QObject.__init__(self)
        self.session = session
        self.signaler = self.session.signaler
        self.main_win = None

        self._process = QProcess()

        if ray.QT_VERSION >= (5, 6):
            self._process.errorOccurred.connect(self._error_in_process)
        self._process.setProcessChannelMode(QProcess.ForwardedChannels)

        self._announce_timer = QTimer()
        self._announce_timer.setInterval(2000)
        self._announce_timer.setSingleShot(True)
        self._announce_timer.timeout.connect(self._announce_timer_out)

        self._port = None
        self._is_announced = False
        self._is_nsm_locked = False

        self.is_local = True
        self.launched_before = False
        self.address = None
        self.url = ''
        self.session_root = ""

        self.signaler.daemon_announce.connect(self._receive_announce)
        self.signaler.daemon_url_changed.connect(self._change_url)

    def _error_in_process(self, error):
        if self.main_win is None:
            return

        self.main_win.daemon_crash()

    def _change_url(self, new_url: str):
        try:
            self.set_osc_address(ray.get_liblo_address(new_url))
        except BaseException:
            return

        self._call_daemon()

    def _call_daemon(self):
        if not self.address:
            # I don't know really why, but it works only with a timer
            QTimer.singleShot(5, self._show_daemon_url_window)
            return

        self._announce_timer.start()

        server = GuiServerThread.instance()
        if not server:
            sys.stderr.write(
                'GUI can not call daemon, GUI OSC server is missing.\n')
            return

        server.announce()

    def _show_daemon_url_window(self):
        self.signaler.daemon_url_request.emit(ErrDaemon.NO_ERROR, self.url)

    def _announce_timer_out(self):
        if self.launched_before:
            self.signaler.daemon_url_request.emit(
                ErrDaemon.NO_ANNOUNCE, self.url)
        else:
            sys.stderr.write(
                _translate(
                    'error',
                    "No announce from ray-daemon. RaySession can't works. Sorry.\n"))
            QApplication.quit()

    def _receive_announce(
            self, src_addr: Address, version: str, server_status: int,
            options: int, session_root: str, is_net_free: int):
        self._announce_timer.stop()

        if version.split('.')[:2] != ray.VERSION.split('.')[:2]:
            # works only if the two firsts digits are the same (ex: 0.6)
            self.signaler.daemon_url_request.emit(
                ErrDaemon.WRONG_VERSION, self.url)
            self.disannounce(src_addr)
            return

        if (CommandLineArgs.net_session_root
                and session_root != CommandLineArgs.net_session_root):
            self.signaler.daemon_url_request.emit(
                ErrDaemon.WRONG_ROOT, self.url)
            self.disannounce(src_addr)
            return

        if not is_net_free:
            self.signaler.daemon_url_request.emit(
                ErrDaemon.FORBIDDEN_ROOT, self.url)
            self.disannounce(src_addr)
            return

        if (CommandLineArgs.out_daemon
                and server_status != ray.ServerStatus.OFF):
            self.signaler.daemon_url_request.emit(ErrDaemon.NOT_OFF, self.url)
            self.disannounce(src_addr)
            return

        self._is_announced = True
        self.address = src_addr
        self._port = src_addr.port
        self.url = src_addr.url
        self.session_root = session_root
        CommandLineArgs.change_session_root(self.session_root)

        self._is_nsm_locked = options & ray.Option.NSM_LOCKED

        if self._is_nsm_locked:
            if self.main_win is not None:
                self.main_win.set_nsm_locked(True)
        elif CommandLineArgs.under_nsm:
            server = GuiServerThread.instance()
            server.to_daemon('/ray/server/set_nsm_locked')

        if self.main_win is not None and self.main_win.waiting_for_patchbay:
            self.main_win.waiting_for_patchbay = False
            server = GuiServerThread.instance()
            server.to_daemon('/ray/server/ask_for_patchbay')

        self.signaler.daemon_announce_ok.emit()
        self.session.set_daemon_options(options)

    def finish_init(self):
        self.main_win = self.session.main_win

    def disannounce(self, address=None):
        if not address:
            address = self.address

        if address:
            server = GuiServerThread.instance()
            server.disannounce(address)

        self._port = None
        self.url = ''
        del self.address
        self.address = None
        self._is_announced = False

    def set_external(self):
        self.launched_before = True

    def set_osc_address(self, address: Address):
        self.address = address
        self.launched_before = True
        self._port = self.address.port
        self.url = self.address.url

        self.is_local = bool(self.address.hostname == socket.gethostname())

    def start(self):
        if self.launched_before:
            self._call_daemon()
            return

        if not CommandLineArgs.force_new_daemon:
            ray_control_process = QProcess()
            ray_control_process.start(
                "ray_control", ['get_port_gui_free', CommandLineArgs.session_root])
            ray_control_process.waitForFinished(2000)

            if ray_control_process.exitCode() == 0:
                port_str_lines = \
                    ray_control_process.readAllStandardOutput().data().decode('utf-8')
                port_str = port_str_lines.partition('\n')[0]

                if port_str and port_str.isdigit():
                    self.address = Address(int(port_str))
                    self._port = self.address.port
                    self.url = self.address.url
                    self.launched_before = True
                    self.is_local = True
                    self._call_daemon()
                    sys.stderr.write(
                        "\033[92m%s\033[0m\n" % (
                            _translate('GUI_daemon',
                                       "Connecting GUI to existing ray-daemon port %i")
                            % self._port))

                    if CommandLineArgs.start_session:
                        server = GuiServerThread.instance()
                        if server:
                            server.send(self.address, '/ray/server/open_session',
                                        CommandLineArgs.start_session)
                    return

        server = GuiServerThread.instance()
        if not server:
            sys.stderr.write(
                "impossible for GUI to launch daemon. server missing.\n")

        # start process
        arguments = ['--gui-url', str(server.url),
                     '--gui-pid', str(os.getpid()),
                     '--osc-port', str(self._port),
                     '--session-root', CommandLineArgs.session_root]

        if CommandLineArgs.start_session:
            arguments.append('--session')
            arguments.append(CommandLineArgs.start_session)

        if CommandLineArgs.debug_only:
            arguments.append('--debug-only')
        elif CommandLineArgs.debug:
            arguments.append('--debug')
        elif CommandLineArgs.no_client_messages:
            arguments.append('--no-client-messages')

        if CommandLineArgs.config_dir:
            arguments.append('--config-dir')
            arguments.append(CommandLineArgs.config_dir)

        self._process.startDetached('ray-daemon', arguments)
        #self.process.start('konsole', ['-e', 'ray-daemon'] + arguments)

    def stop(self):
        if self.launched_before:
            self.disannounce()
            QTimer.singleShot(10, QApplication.quit)
            return

        server = GuiServerThread.instance()
        server.to_daemon('/ray/server/quit')
        QTimer.singleShot(50, QApplication.quit)

    def set_new_osc_address(self):
        if not (self.address or self._port):
            self._port = ray.get_free_osc_port()
            self.address = Address(self._port)

    def is_announced(self):
        return self._is_announced

    def get_port(self):
        return self._port



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# PatchBay Canvas engine using QGraphicsView/Scene
# Copyright (C) 2010-2019 Filipe Coelho <falktx@falktx.com>
# Copyright (C) 2019-2022 Mathieu Picot <picotmathieu@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the doc/GPL.txt file.

from typing import TYPE_CHECKING, Iterator, Optional, Union
from enum import IntEnum, IntFlag

from PyQt5.QtCore import QPointF, QRectF, QSettings, QPoint
from PyQt5.QtWidgets import QGraphicsItem

if TYPE_CHECKING:
    # all these classes are not importable normally because
    # it would make a circular import
    # they are imported only from IDE to get types
    from .theme import Theme
    from .theme_manager import ThemeManager
    from .scene import PatchScene
    from .box_widget import BoxWidget
    from .port_widget import PortWidget
    from .portgroup_widget import PortgroupWidget
    from .line_widget import LineWidget
    from .patchcanvas import CanvasObject


# Maximum Id for a plugin, treated as invalid/zero if above this value
MAX_PLUGIN_ID_ALLOWED = 0x7FF


class PortMode(IntFlag):
    NULL = 0x00
    INPUT = 0x01
    OUTPUT = 0x02
    BOTH = INPUT | OUTPUT
    
    def opposite(self) -> 'PortMode':
        if self is self.INPUT:
            return self.OUTPUT
        if self is self.OUTPUT:
            return self.INPUT
        if self is self.BOTH:
            return self.NULL
        if self is self.NULL:
            return self.BOTH


class PortType(IntFlag):
    NULL = 0x00
    AUDIO_JACK = 0x01
    MIDI_JACK = 0x02
    MIDI_ALSA = 0x04
    VIDEO = 0x08
    PARAMETER = 0x10


class PortSubType(IntFlag):
    ''' a2j ports are MIDI ports, we only specify a decoration for them.
        CV ports are audio ports, but we prevent to connect an output CV port
        to a regular audio port to avoid material destruction, CV ports also
        look different, simply because this is absolutely not the same use.'''
    REGULAR = 0x01
    CV = 0x02
    A2J = 0x04


# Callback Actions
class CallbackAct(IntEnum):
    GROUP_INFO = 0          # group_id: int
    GROUP_RENAME = 1        # group_id: int
    GROUP_SPLIT = 2         # group_id: int
    GROUP_JOIN = 3          # group_id: int
    GROUP_JOINED = 4        # group_id: int
    GROUP_MOVE = 5          # group_id: int, splitted_mode: PortMode, x: int, y: int
    GROUP_WRAP = 6          # group_id: int, folded: bool
    GROUP_LAYOUT_CHANGE = 7 # group_id: int, layout_mode: BoxLayoutMode, splitted_mode: PortMode
    GROUP_SELECTED = 8      # group_id: int, splitted_mode: PortMode
    PORTGROUP_ADD = 9       # group_id: int, portgrp_id: int, port_mode: PortMode,
                            #      port_type: PortType, port_ids: tuple[int]
    PORTGROUP_REMOVE = 10   # group_id: int, portgrp_id: int
    PORT_INFO = 11          # group_id: int, port_id: int
    PORT_RENAME = 12        # group_id: int, port_id: int
    PORTS_CONNECT = 13      # group_out_id: int, port_out_id: int,
                            #      group_in_id: int, port_in_id: int
    PORTS_DISCONNECT = 14   # conn_id: int
    PORT_MENU_CALL = 15     # group_id: int, port_id: int, connect_only: bool,
                            #      x_pos: int, y_pos: int
    PLUGIN_CLONE = 16       # plugin_id: int
    PLUGIN_EDIT = 17        # plugin_id: int
    PLUGIN_RENAME = 18      # plugin_id: int
    PLUGIN_REPLACE = 19     # plugin_id: int
    PLUGIN_REMOVE = 20      # plugin_id: int
    PLUGIN_SHOW_UI = 21     # plugin_id: int
    INLINE_DISPLAY = 22     # plugin_id: int
    BG_RIGHT_CLICK = 23     # 
    BG_DOUBLE_CLICK = 24    # 
    CLIENT_SHOW_GUI = 25    # group_id: int, visible: bool
    THEME_CHANGED = 26      # theme_name: str


class BoxType(IntEnum):
    APPLICATION = 0
    HARDWARE = 1
    MONITOR = 2
    DISTRHO = 3
    FILE = 4
    PLUGIN = 5
    LADISH_ROOM = 6
    CLIENT = 7
    INTERNAL = 8


# Split Option
class BoxSplitMode(IntEnum):
    UNDEF = 0
    NO = 1
    YES = 2


# define the way ports are put in a box
class BoxLayoutMode(IntEnum):
    AUTO = 0
    HIGH = 1
    LARGE = 2


# For Repulsive boxes
class Direction(IntEnum):
    NONE = 0
    LEFT = 1
    RIGHT = 2
    UP = 3
    DOWN = 4


class CanvasItemType(IntEnum):
    # this enum is still here if really needed
    # but never really used.
    # Prefer use isinstance(item, type) if possible
    # because IDE will know easier with which class
    # we are dealing.
    BOX = QGraphicsItem.UserType + 1
    ICON = QGraphicsItem.UserType + 2
    PORT = QGraphicsItem.UserType + 3
    PORTGROUP = QGraphicsItem.UserType + 4
    BEZIER_LINE = QGraphicsItem.UserType + 5
    BEZIER_LINE_MOV = QGraphicsItem.UserType + 6
    RUBBERBAND = QGraphicsItem.UserType + 7


# Canvas options
class CanvasOptionsObject:
    theme_name = ""
    auto_hide_groups = True
    auto_select_items = False
    show_shadows = False
    inline_displays = 0
    elastic = True
    borders_navigation = True
    prevent_overlap = True
    max_port_width = 180
    semi_hide_opacity = 0.20
    default_zoom = 100
    box_grouped_auto_layout_ratio = 1.0


# Canvas features
class CanvasFeaturesObject:
    group_info = False
    group_rename = False
    port_info = True
    port_rename = False
    handle_group_pos = False


# ------------------------

# object lists            
class GroupObject:
    group_id: int
    group_name: str
    split: int
    box_type: int
    icon_name: str
    layout_modes: dict[PortMode, BoxLayoutMode]
    plugin_id: int
    plugin_ui: int # to verify
    plugin_inline: int # to verify
    null_pos: QPoint
    in_pos: QPoint
    out_pos: QPoint
    handle_client_gui: bool
    gui_visible: bool
    widgets: list
    if TYPE_CHECKING:
        widgets: list[Optional[BoxWidget]]

    def copy_no_widget(self) -> 'GroupObject':
        group_copy = GroupObject()
        group_copy.__dict__ = self.__dict__.copy()
        group_copy.widgets = [None, None]
        return group_copy


class ConnectableObject:
    group_id: int
    port_mode: PortMode
    port_type: PortType
    port_subtype: PortSubType
    portgrp_id: int
    
    def get_port_ids(self) -> tuple[int]:
        return ()


class PortObject(ConnectableObject):
    port_id: int
    port_name: str
    widget: object
    portgrp: object
    if TYPE_CHECKING:
        widget: PortWidget
        portgrp: 'PortgrpObject'

    pg_pos = 0 # index in the portgroup (if any)
    pg_len = 1 # length of the portgroup (if any)

    def __repr__(self) -> str:
        return f"PortObject({self.port_name} - {self.port_id} - {self.portgrp_id})"

    def copy_no_widget(self) -> 'PortObject':
        port_copy = PortObject()
        port_copy.__dict__ = self.__dict__.copy()
        port_copy.widget = None
        return port_copy

    def get_port_ids(self) -> tuple[int]:
        return (self.port_id,)

    def set_portgroup_id(self, pg_id: int, pg_pos: int, pg_len: int):
        self.portgrp_id = pg_id
        self.pg_pos = pg_pos
        self.pg_len = pg_len
        if self.widget is not None:
            self.widget.set_portgroup_id(pg_id, pg_pos, pg_len)


class PortgrpObject(ConnectableObject):
    port_id_list: list[int]
    widget: object
    if TYPE_CHECKING:
        widget: PortgroupWidget

    def __init__(self):
        self.ports = list[PortObject]()

    def copy_no_widget(self) -> 'PortObject':
        portgrp_copy = PortgrpObject()
        portgrp_copy.__dict__ = self.__dict__.copy()
        portgrp_copy.widget = None
        return portgrp_copy
    
    def get_port_ids(self) -> tuple[int]:
        return tuple(self.port_id_list)


class ConnectionObject:
    connection_id: int
    group_in_id: int
    port_in_id: int
    group_out_id: int
    port_out_id: int
    widget: object
    if TYPE_CHECKING:
        widget: LineWidget

    def copy_no_widget(self) -> 'ConnectionObject':
        conn_copy = ConnectionObject()
        conn_copy.__dict__ = self.__dict__.copy()
        conn_copy.widget = None
        return conn_copy

    def matches(self, group_id_1: int, port_ids_list_1: list[int],
                group_id_2: int, port_ids_list_2: list[int]) -> bool:
        if (self.group_in_id == group_id_1
            and self.port_in_id in port_ids_list_1
            and self.group_out_id == group_id_2
            and self.port_out_id in port_ids_list_2):
                return True
        elif (self.group_in_id == group_id_2
            and self.port_in_id in port_ids_list_2
            and self.group_out_id == group_id_1
            and self.port_out_id in port_ids_list_1):
                return True
        else:
            return False

    def concerns(self, group_id: int, port_ids_list: list[int]) -> bool:
        if (self.group_in_id == group_id
                and self.port_in_id in port_ids_list):
            return True
        elif (self.group_out_id == group_id
            and self.port_out_id in port_ids_list):
            return True
        else:
            return False


class ClipboardElement:
    port_type: PortType
    port_mode: PortMode
    group_id: int
    port_id: int
    group_port_ids: list[int]


# Main Canvas object
class Canvas:
    def __init__(self):
        self.qobject = None
        self.settings = None
        self.theme = None

        self.initiated = False
        self.theme_paths = ()
        self.theme_manager = None

        self.group_list = list[GroupObject]()
        self.port_list = list[PortObject]()
        self.portgrp_list = list[PortgrpObject]()
        self.connection_list = list[ConnectionObject]()

        self._groups_dict = dict[int, GroupObject]()
        self._all_boxes = []
        self._ports_dict = dict[int, dict[int, PortObject]]()
        self._portgrps_dict = dict[int, dict[int, PortgrpObject]]()
        self._conns_dict = dict[int, ConnectionObject]()
        self._conns_outin_dict = dict[int, dict[int, dict[int, ConnectionObject]]]()
        self._conns_inout_dict = dict[int, dict[int, dict[int, ConnectionObject]]]()

        self.clipboard = list[ClipboardElement]()
        self.clipboard_cut = True
        self.group_plugin_map = {}

        self.scene = None
        self.last_z_value = 0
        self.initial_pos = QPointF(0, 0)
        self.size_rect = QRectF()

        self.is_line_mov = False
        self.loading_items = False
        self.menu_shown = False
        self.menu_click_pos = QPoint(0, 0)
        
        # This is only to get object methods in IDE everywhere.
        if TYPE_CHECKING:
            self.qobject = CanvasObject()
            self.theme = Theme()
            self.theme_manager = ThemeManager()
            self.scene = PatchScene()
            self.settings = QSettings()
            self.options = CanvasOptionsObject()
            self.features = CanvasFeaturesObject()
            
            self._all_boxes = list[BoxWidget]()

    def callback(self, action: CallbackAct, value1: int,
                 value2: int, value_str: str):
        # has to be redefined in patchcanvas.init()
        pass
    
    def clear_all(self):
        self.port_list.clear()
        self._ports_dict.clear()
        self.portgrp_list.clear()
        self._ports_dict.clear()
        self.connection_list.clear()
        self._conns_dict.clear()
        self._conns_outin_dict.clear()
        self._conns_inout_dict.clear()

    def add_group(self, group: GroupObject):
        self.group_list.append(group)
        self._groups_dict[group.group_id] = group
        for widget in group.widgets:
            if widget is not None:
                self._all_boxes.append(widget)

    def add_port(self, port: PortObject):
        self.port_list.append(port)

        gp_dict = self._ports_dict.get(port.group_id)
        if gp_dict is None:
            self._ports_dict[port.group_id] = {port.port_id: port}
        else:
            self._ports_dict[port.group_id][port.port_id] = port
    
    def add_portgroup(self, portgrp: PortgrpObject):
        self.portgrp_list.append(portgrp)

        gp_dict = self._portgrps_dict.get(portgrp.group_id)
        if gp_dict is None:
            self._portgrps_dict[portgrp.group_id] = {portgrp.portgrp_id: portgrp}
        else:
            self._portgrps_dict[portgrp.group_id][portgrp.portgrp_id] = portgrp
        
        for port in self.list_ports(group_id=portgrp.group_id):
            if port.port_id in portgrp.port_id_list:
                portgrp.ports.append(port)
                port.portgrp = portgrp
    
    def add_connection(self, conn: ConnectionObject):
        self.connection_list.append(conn)
        self._conns_dict[conn.connection_id] = conn

        gp_outin_out_dict = self._conns_outin_dict.get(conn.group_out_id)
        gp_inout_in_dict = self._conns_inout_dict.get(conn.group_in_id)
        
        if gp_outin_out_dict is None:
            self._conns_outin_dict[conn.group_out_id] = {
                conn.group_in_id: {conn.connection_id: conn}}
        else:
            if conn.group_in_id in gp_outin_out_dict:
                gp_outin_out_dict[conn.group_in_id][conn.connection_id] = conn
            else:
                gp_outin_out_dict[conn.group_in_id] = {conn.connection_id: conn}
        
        if gp_inout_in_dict is None:
            self._conns_inout_dict[conn.group_in_id] = {
                conn.group_out_id: {conn.connection_id: conn}}
        else:
            if conn.group_out_id in gp_inout_in_dict:
                gp_inout_in_dict[conn.group_out_id][conn.connection_id] = conn
            else:
                gp_inout_in_dict[conn.group_out_id] = {conn.connection_id: conn}
    
    def remove_group(self, group: GroupObject):
        self.group_list.remove(group)
        if group.group_id in self._groups_dict.keys():
            self._groups_dict.pop(group.group_id)
        
        for widget in group.widgets:
            if widget is not None and widget in self._all_boxes:
                self._all_boxes.remove(widget)
    
    def remove_port(self, port: PortObject):
        if port in self.port_list:
            self.port_list.remove(port)
        
        gp_dict = self._ports_dict.get(port.group_id)
        if gp_dict and port.port_id in gp_dict.keys():
            gp_dict.pop(port.port_id)
    
    def remove_portgroup(self, portgrp: PortgrpObject):
        if portgrp in self.portgrp_list:
            self.portgrp_list.remove(portgrp)
        
        gp_dict = self._portgrps_dict.get(portgrp.group_id)
        if gp_dict and portgrp.portgrp_id in gp_dict.keys():
            gp_dict.pop(portgrp.portgrp_id)
            
        for port in self.list_ports(group_id=portgrp.group_id):
            if port.port_id in portgrp.port_id_list:
                port.portgrp = None
    
    def remove_connection(self, conn: ConnectionObject):
        if conn in self.connection_list:
            self.connection_list.remove(conn)
            
        try:
            self._conns_dict.pop(conn.connection_id)
            self._conns_outin_dict[conn.group_out_id][conn.group_in_id].pop(
                conn.connection_id)
            self._conns_inout_dict[conn.group_in_id][conn.group_out_id].pop(
                conn.connection_id)
        except:
            pass
    
    def get_group(self, group_id: int) -> GroupObject:
        return self._groups_dict.get(group_id)
    
    def get_port(self, group_id: int, port_id: int) -> PortObject:
        gp = self._ports_dict.get(group_id)
        if gp is None:
            return None
        return gp.get(port_id)

    def get_portgroup(self, group_id: int, portgrp_id: int) -> PortgrpObject:
        gp = self._portgrps_dict.get(group_id)
        if gp is None:
            return None
        return gp.get(portgrp_id)

    def get_connection(self, connection_id: int) -> ConnectionObject:
        if connection_id in self._conns_dict.keys():
            return self._conns_dict[connection_id]

    def list_boxes(self) -> list['BoxWidget']:
        return self._all_boxes

    def list_ports(self, group_id=None) -> Iterator[PortObject]:
        if group_id is None:
            for port in self.port_list:
                yield port
            return     
        
        if group_id in self._ports_dict.keys():
            for port in self._ports_dict[group_id].values():
                yield port
            
    def list_portgroups(self, group_id=None) -> Iterator[PortgrpObject]:
        if group_id is None:
            for portgrp in self.portgrp_list:
                yield portgrp
            return     
        
        if group_id in self._portgrps_dict.keys():
            for portgrp in self._portgrps_dict[group_id].values():
                yield portgrp
                
    def list_connections(
            self, *connectables: ConnectableObject,
            group_in_id=None, group_out_id=None,
            group_id=None) -> Iterator[ConnectionObject]:
        if (not connectables
                and group_id is None
                and group_in_id is None
                and group_out_id is None):
            # no filter, list all connections
            for conn in self._conns_dict.values():
                yield conn
            return
        
        if len(connectables) > 2:
            return
        
        port_out_ids, port_in_ids = tuple[int](), tuple[int]()
        
        # check infos from connectables (port/portgroup)
        if len(connectables) == 2:
            if (connectables[0].port_mode
                    is not connectables[1].port_mode.opposite()):
                return
            
            if connectables[0].port_mode is PortMode.OUTPUT:
                connectable_out, connectable_in = connectables
            else:
                connectable_in, connectable_out = connectables
            
            group_out_id = connectable_out.group_id
            group_in_id = connectable_in.group_id
            port_out_ids = connectable_out.get_port_ids()
            port_in_ids = connectable_in.get_port_ids()

        elif len(connectables) == 1:
            if connectables[0].port_mode is PortMode.OUTPUT:
                group_out_id = connectables[0].group_id
                port_out_ids = connectables[0].get_port_ids()
            else:
                group_in_id = connectables[0].group_id
                port_in_ids = connectables[0].get_port_ids()

        # check if group_id should be group_out_id or group_in_id
        if (group_id is not None
                and (group_out_id is not None or group_in_id is not None)):
            if group_out_id is None:
                group_out_id = group_id
            elif group_in_id is None:
                group_in_id = group_id
            group_id = None

        # no precision for group_id
        # we check first in the out dict, then in the in dict
        if group_id is not None:
            gp_out = self._conns_outin_dict.get(group_id)
            if gp_out is not None:
                for gp_out_in in gp_out.values():
                    for conn in gp_out_in.values():
                        yield conn
        
            gp_in = self._conns_inout_dict.get(group_id)
            if gp_in is not None:
                for gp_id, gp_in_out in gp_in.items():
                    if gp_id == group_id:
                        # connections here have already been yielded
                        continue
                    for conn in gp_in_out.values():
                        yield conn
            return
        
        # here we are sure we have group_out_id or group_in_id filter(s)
        if group_out_id is not None:
            gp_out = self._conns_outin_dict.get(group_out_id)
            if gp_out is None:
                return
            
            if group_in_id is not None:
                gp_in = gp_out.get(group_in_id)
                if gp_in is None:
                    return
                
                for conn in gp_in.values():
                    if ((not port_out_ids or conn.port_out_id in port_out_ids)
                            and (not port_in_ids or conn.port_in_id in port_in_ids)):
                        yield conn
                return
            
            for gp_in in gp_out.values():
                for conn in gp_in.values():
                    if not port_out_ids or conn.port_out_id in port_out_ids:
                        yield conn
            return

        # here we are sure we have no group_out_id filter
        gp_in = self._conns_inout_dict.get(group_in_id)
        if gp_in is None:
            return
        
        for gp_out in gp_in.values():
            for conn in gp_out.values():
                if not port_in_ids or conn.port_in_id in port_in_ids:
                    yield conn

# -----------------------------

# Global objects
canvas = Canvas()

options = CanvasOptionsObject()
features = CanvasFeaturesObject()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# PatchBay Canvas engine using QGraphicsView/Scene
# Copyright (C) 2010-2019 Filipe Coelho <falktx@falktx.com>
# Copyright (C) 2019-2022 Mathieu Picot <picotmathieu@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the doc/GPL.txt file.

# global imports
import logging
from pathlib import Path
from typing import Callable, Optional
from PyQt5.QtCore import (pyqtSlot, QObject, QPoint, QPointF, QRectF,
                          QSettings, QTimer, pyqtSignal)
from PyQt5.QtWidgets import QAction, QMainWindow

# local imports
from .init_values import (
    CanvasItemType,
    PortSubType,
    PortType,
    canvas,
    options,
    features,
    CanvasOptionsObject,
    CanvasFeaturesObject,
    CallbackAct,
    MAX_PLUGIN_ID_ALLOWED,
    GroupObject,
    PortObject,
    PortgrpObject,
    ConnectionObject,
    PortMode,
    BoxSplitMode,
    BoxLayoutMode,
    BoxType,
)

from .utils import canvas_callback, get_new_group_pos
from .box_widget import BoxWidget
from .port_widget import PortWidget
from .line_widget import LineWidget
from .theme_manager import ThemeManager
from .scene import PatchScene
from .scene_view import PatchGraphicsView

_logger = logging.getLogger(__name__)
# used by patchbay_api decorator to get function_name
# and arguments, easily usable by logger
_logging_str = ''

# decorator
def patchbay_api(func):
    ''' decorator for API callable functions.
        It makes debug logs and also a global logging string
        usable directly in the functions'''
    def wrapper(*args, **kwargs):
        args_strs = [str(arg) for arg in args]
        args_strs += [f"{k}={v}" for k, v in kwargs.items()]

        global _logging_str
        _logging_str = f"{func.__name__}({', '.join(args_strs)})"
        _logger.debug(_logging_str)
        return func(*args, **kwargs)
    return wrapper


class CanvasObject(QObject):
    port_added = pyqtSignal(int, int)
    port_removed = pyqtSignal(int, int)
    connection_added = pyqtSignal(int)
    connection_removed = pyqtSignal(int)
    move_boxes_finished = pyqtSignal()

    def __init__(self, parent=None):
        QObject.__init__(self, parent)
        self.groups_to_join = list[int]()
        self.move_boxes_finished.connect(self.join_after_move)

    @pyqtSlot()
    def port_context_menu_disconnect(self):
        try:
            con_ids_list = list(self.sender().data())
        except:
            return

        for connection_id in con_ids_list:
            if type(connection_id) != int:
                continue

            canvas_callback(CallbackAct.PORTS_DISCONNECT, connection_id)

    @pyqtSlot()
    def portgroup_the_ports(self):
        try:
            sender = self.sender()
            assert isinstance(sender, QAction)
            all_data = sender.data()
            assert isinstance(all_data, tuple) and len(all_data) == 2
        except:
            return
        
        all_data: tuple[PortObject, PortObject]
        
        canvas_callback(
            CallbackAct.PORTGROUP_ADD,
            all_data[0].group_id, all_data[0].port_mode, all_data[0].port_type,
            tuple([p.port_id for p in all_data]))

    def join_after_move(self):
        for group_id in self.groups_to_join:
            join_group(group_id)

        self.groups_to_join.clear()


def _get_stored_canvas_position(key, fallback_pos):
    try:
        return canvas.settings.value(
            "CanvasPositions/" + key, fallback_pos, type=QPointF)
    except:
        return fallback_pos




@patchbay_api
def init(view: PatchGraphicsView, callback: Callable,
          theme_paths: tuple[Path], fallback_theme: str):
    if canvas.initiated:
        _logger.critical("init() - already initiated")
        return

    if not callback:
        _logger.critical("init() - fatal error: callback not set")
        return

    canvas.callback = callback
    canvas.scene = PatchScene(view)
    view.setScene(canvas.scene)
    
    canvas.last_z_value = 0
    canvas.initial_pos = QPointF(0, 0)
    canvas.size_rect = QRectF()

    if canvas.qobject is None:
        canvas.qobject = CanvasObject()

    if canvas.settings is None:
        # TODO : may remove this because it is not used
        # while features.handle_positions is False. 
        canvas.settings = QSettings()

    if canvas.theme_manager is None:
        canvas.theme_manager = ThemeManager(theme_paths)
        if not canvas.theme_manager.set_theme(options.theme_name):
            if canvas.theme_manager.set_theme(fallback_theme):
                _logger.warning(
                f"theme {options.theme_name}' has not been found,"
                f"use '{fallback_theme}' instead.")
            else:
                _logger.warning(
                f"theme '{options.theme_name}' has not been found,"
                "use the very ugly fallback theme.")
                canvas.theme_manager.set_fallback_theme()

        canvas.theme.load_cache()

    canvas.scene.zoom_reset()
    canvas.initiated = True

@patchbay_api
def clear():
    if not canvas.initiated:
        return
    
    group_list_ids = [g.group_id for g in canvas.group_list]
    port_list_ids = [(p.group_id, p.port_id) for p in canvas.list_ports()]
    connection_list_ids = [c.connection_id for c in canvas.list_connections()]

    for idx in connection_list_ids:
        disconnect_ports(idx)

    for group_id, port_id in port_list_ids:
        remove_port(group_id, port_id)

    for idx in group_list_ids:
        remove_group(idx)

    canvas.last_z_value = 0

    canvas.clear_all()
    canvas.group_plugin_map = {}

    canvas.scene.clearSelection()

    for item in canvas.scene.items():
        if item.type() in (CanvasItemType.ICON, CanvasItemType.RUBBERBAND):
            continue
        canvas.scene.removeItem(item)
        del item

    canvas.initiated = False

    QTimer.singleShot(0, canvas.scene.update)

# ------------------------------------------------------------------------------------------------------------
@patchbay_api
def set_initial_pos(x: int, y: int):
    canvas.initial_pos.setX(x)
    canvas.initial_pos.setY(y)

@patchbay_api
def set_canvas_size(x: int, y: int, width: int, height: int):
    canvas.size_rect.setX(x)
    canvas.size_rect.setY(y)
    canvas.size_rect.setWidth(width)
    canvas.size_rect.setHeight(height)
    canvas.scene.update_limits()
    canvas.scene.fix_scale_factor()

@patchbay_api
def set_loading_items(yesno: bool):
    '''while canvas is loading items (groups or ports, connections...)
    then, items will be added, but not redrawn.
    This is an optimization that prevents a lot of redraws.
    Think to set loading items at False and use redraw_all_groups
    or redraw_group once the long operation is finished'''
    canvas.loading_items = yesno

@patchbay_api
def add_group(group_id: int, group_name: str, split=BoxSplitMode.UNDEF,
              icon_type=BoxType.APPLICATION, icon_name='', layout_modes={},
              null_xy=(0, 0), in_xy=(0, 0), out_xy=(0, 0),
              split_animated=False):
    if canvas.get_group(group_id) is not None:
        _logger.error(f"{_logging_str} - group already exists.")
        return

    if split is BoxSplitMode.UNDEF and icon_type is BoxType.HARDWARE:
        split = BoxSplitMode.YES

    group_box = BoxWidget(group_id, group_name, icon_type, icon_name)
    
    group = GroupObject()
    group.group_id = group_id
    group.group_name = group_name
    group.split = bool(split == BoxSplitMode.YES)
    group.box_type = icon_type
    group.icon_name = icon_name
    group.layout_modes = layout_modes
    group.plugin_id = -1
    group.plugin_ui = False
    group.plugin_inline = False
    group.handle_client_gui = False
    group.gui_visible = False
    group.null_pos = QPoint(*null_xy)
    group.in_pos = QPoint(*in_xy)
    group.out_pos = QPoint(*out_xy)
    group.widgets = list[BoxWidget]()
    group.widgets.append(group_box)
    group.widgets.append(None)

    if split == BoxSplitMode.YES:
        group_box.set_split(True, PortMode.OUTPUT)

        if features.handle_group_pos:
            new_pos = _get_stored_canvas_position(
                group_name + "_OUTPUT", get_new_group_pos(False))
            canvas.scene.add_box_to_animation(group_box, new_pos.x(), new_pos.y())
        else:
            if split_animated:
                group_box.setPos(group.null_pos)
            else:
                group_box.setPos(group.out_pos)
            
        group_sbox = BoxWidget(group_id, group_name, icon_type, icon_name)
        group_sbox.set_split(True, PortMode.INPUT)

        group.widgets[1] = group_sbox

        if features.handle_group_pos:
            new_pos = _get_stored_canvas_position(
                group_name + "_INPUT", get_new_group_pos(True))
            canvas.scene.add_box_to_animation(group_sbox, new_pos.x(), new_pos.y())
        else:
            if split_animated:
                group_sbox.setPos(group.null_pos)
            else:
                group_sbox.setPos(group.in_pos)

        canvas.last_z_value += 1
        group_sbox.setZValue(canvas.last_z_value)

    else:
        group_box.set_split(False)

        if features.handle_group_pos:
            group_box.setPos(_get_stored_canvas_position(
                group_name, get_new_group_pos(False)))
        else:
            # Special ladish fake-split groups
            group_box.setPos(group.null_pos)


    canvas.last_z_value += 1
    group_box.setZValue(canvas.last_z_value)
    canvas.add_group(group)

    if canvas.loading_items:
        return

    if split_animated:
        for box in group.widgets:
            if box is not None:
                if box.get_splitted_mode() is PortMode.OUTPUT:
                    canvas.scene.add_box_to_animation(
                        box, group.out_pos.x(), group.out_pos.y())
                elif box.get_splitted_mode() is PortMode.INPUT:
                    canvas.scene.add_box_to_animation(
                        box, group.in_pos.x(), group.in_pos.y())

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def remove_group(group_id: int, save_positions=True):
    group = canvas.get_group(group_id)
    if group is None:
        _logger.error(f"{_logging_str} - unable to find group to remove")
        return
              
    item = group.widgets[0]
    group_name = group.group_name

    if group.split:
        s_item = group.widgets[1]

        if features.handle_group_pos and save_positions:
            canvas.settings.setValue(
                "CanvasPositions/%s_OUTPUT" % group_name, item.pos())
            canvas.settings.setValue(
                "CanvasPositions/%s_INPUT" % group_name, s_item.pos())
            canvas.settings.setValue(
                "CanvasPositions/%s_SPLIT" % group_name, BoxSplitMode.YES)

        for moving_box in canvas.scene.move_boxes:
            if moving_box.widget is s_item:
                canvas.scene.move_boxes.remove(moving_box)
                break

        s_item.remove_icon_from_scene()
        
        canvas.scene.removeItem(s_item)
        del s_item

    else:
        if features.handle_group_pos and save_positions:
            canvas.settings.setValue(
                "CanvasPositions/%s" % group_name, item.pos())
            canvas.settings.setValue(
                "CanvasPositions/%s_SPLIT" % group_name, BoxSplitMode.NO)
    
    for moving_box in canvas.scene.move_boxes:
        if moving_box.widget is item:
            canvas.scene.move_boxes.remove(moving_box)
            break

    item.remove_icon_from_scene()
    canvas.scene.removeItem(item)
    del item

    canvas.remove_group(group)
    canvas.group_plugin_map.pop(group.plugin_id, None)

    if canvas.loading_items:
        return

    QTimer.singleShot(0, canvas.scene.update)
    QTimer.singleShot(0, canvas.scene.resize_the_scene)

@patchbay_api
def rename_group(group_id: int, new_group_name: str):
    group = canvas.get_group(group_id)
    if group is None:
        _logger.critical(f"{_logging_str} - unable to find group to rename")
        return

    group.group_name = new_group_name
    group.widgets[0].set_group_name(new_group_name)

    if group.split and group.widgets[1]:
        group.widgets[1].set_group_name(new_group_name)

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def split_group(group_id: int, on_place=False):
    item = None

    # Step 1 - Store all Item data
    group = canvas.get_group(group_id)
    if group is None:
        _logger.error(f"{_logging_str} - unable to find group to split")
        return
    
    if group.split:
        _logger.error(
            f"{_logging_str} - group is already splitted")
        return

    item = group.widgets[0]
    tmp_group = group.copy_no_widget()
        
    if on_place and item is not None:
        pos = item.pos()
        rect = item.boundingRect()
        y = int(pos.y())
        x = int(pos.x())
        tmp_group.in_pos = QPoint(x - int(rect.width() / 2), y)
        tmp_group.out_pos = QPoint(x + int(rect.width() / 2), y)

    wrap = item.is_wrapped()

    portgrps_data = [pg.copy_no_widget() for pg 
                     in canvas.list_portgroups(group_id=group_id)]
    ports_data = [p.copy_no_widget() for p in canvas.list_ports(group_id=group_id)]
    conns_data = [c.copy_no_widget()
                  for c in canvas.list_connections(group_id=group_id)]

    canvas.loading_items = True

    # Step 2 - Remove Item and Children
    for conn in conns_data:
        disconnect_ports(conn.connection_id)

    for portgrp in portgrps_data:
        if portgrp.group_id == group_id:
            remove_portgroup(group_id, portgrp.portgrp_id)

    for port in ports_data:
        if port.group_id == group_id:
            remove_port(group_id, port.port_id)

    remove_group(group_id)

    g = tmp_group

    # Step 3 - Re-create Item, now split
    add_group(group_id, g.group_name, BoxSplitMode.YES,
              g.box_type, g.icon_name, g.layout_modes,
              null_xy=(g.null_pos.x(), g.null_pos.y()),
              in_xy=(g.in_pos.x(), g.in_pos.y()),
              out_xy=(g.out_pos.x(), g.out_pos.y()),
              split_animated=True)

    if g.handle_client_gui:
        set_optional_gui_state(group_id, g.gui_visible)

    if g.plugin_id >= 0:
        set_group_as_plugin(group_id, g.plugin_id, g.plugin_ui, g.plugin_inline)

    for port in ports_data:
        add_port(group_id, port.port_id, port.port_name, port.port_mode,
                 port.port_type, port.port_subtype)

    for portgrp in portgrps_data:
        add_portgroup(group_id, portgrp.portgrp_id, portgrp.port_mode,
                      portgrp.port_type, portgrp.port_subtype,
                      portgrp.port_id_list)

    for conn in conns_data:
        connect_ports(conn.connection_id, conn.group_out_id, conn.port_out_id,
                      conn.group_in_id, conn.port_in_id)

    canvas.loading_items = False
    
    for box in canvas.get_group(group_id).widgets:
        if box is not None:
            box.set_wrapped(wrap, animate=False)
            box.update_positions(even_animated=True)
            
            if box.get_current_port_mode() is PortMode.OUTPUT:
                canvas.scene.add_box_to_animation(box, g.out_pos.x(), g.out_pos.y())
            else:
                canvas.scene.add_box_to_animation(box, g.in_pos.x(), g.in_pos.y())

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def join_group(group_id: int):
    item = None
    s_item = None

    # Step 1 - Store all Item data
    group = canvas.get_group(group_id)
    if group is None:
        _logger.error(f"{_logging_str} - unable to find groups to join")
        return

    if not group.split:
        _logger.error(f"{_logging_str} - group is not splitted")
        return

    item, s_item = group.widgets
    tmp_group = group.copy_no_widget()

    wrap = item.is_wrapped() and s_item.is_wrapped()

    portgrps_data = [pg.copy_no_widget() for pg in
                     canvas.list_portgroups(group_id=group_id)]
    ports_data = [p.copy_no_widget()
                  for p in canvas.list_ports(group_id=group_id)]
    conns_data = [c.copy_no_widget() 
                  for c in canvas.list_connections(group_id=group_id)]

    canvas.loading_items = True

    # Step 2 - Remove Item and Children
    for conn in conns_data:
        disconnect_ports(conn.connection_id)

    for portgrp in portgrps_data:
        remove_portgroup(group_id, portgrp.portgrp_id)

    for port in ports_data:
        remove_port(group_id, port.port_id)

    remove_group(group_id, save_positions=False)

    g = tmp_group
    # Step 3 - Re-create Item, now together
    add_group(group_id, g.group_name, BoxSplitMode.NO,
              g.box_type, g.icon_name, g.layout_modes,
              null_xy=(g.null_pos.x(), g.null_pos.y()),
              in_xy=(g.in_pos.x(), g.in_pos.y()),
              out_xy=(g.out_pos.x(), g.out_pos.y()))

    if g.handle_client_gui:
        set_optional_gui_state(group_id, g.gui_visible)

    if g.plugin_id >= 0:
        set_group_as_plugin(group_id, g.plugin_id, g.plugin_ui, g.plugin_inline)

    for port in ports_data:
        add_port(group_id, port.port_id, port.port_name, port.port_mode,
                 port.port_type, port.port_subtype)

    for portgrp in portgrps_data:
        add_portgroup(group_id, portgrp.portgrp_id, portgrp.port_mode,
                      portgrp.port_type, portgrp.port_subtype,
                      portgrp.port_id_list)

    for conn in conns_data:
        connect_ports(conn.connection_id, conn.group_out_id, conn.port_out_id,
                      conn.group_in_id, conn.port_in_id)

    for box in canvas.get_group(group_id).widgets:
        if box is not None:
            box.set_wrapped(wrap, animate=False)

    canvas.loading_items = False
    redraw_group(group_id)

    canvas.callback(CallbackAct.GROUP_JOINED, group_id)

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def repulse_all_boxes():
    if options.prevent_overlap:
        canvas.scene.full_repulse()

@patchbay_api
def redraw_all_groups(force_no_prevent_overlap=False):    
    # We are redrawing all groups.
    # For optimization reason we prevent here to resize the scene
    # at each group draw, we'll do it once all is done,
    # same for prevent_overlap.
    elastic = options.elastic
    prevent_overlap = options.prevent_overlap
    options.elastic = False
    options.prevent_overlap = False
        
    for box in canvas.list_boxes():
        box.update_positions(without_connections=True)
    
    for connection in canvas.list_connections():
        if connection.widget is not None:
            connection.widget.update_line_pos()
    
    if canvas.scene is None:
        options.elastic = elastic
        options.prevent_overlap = prevent_overlap
        return
    
    if elastic:
        canvas.scene.set_elastic(True)
    
    if prevent_overlap:
        options.prevent_overlap = True
        if not force_no_prevent_overlap:
            repulse_all_boxes()
    
    if not elastic or prevent_overlap:
        QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def redraw_group(group_id: int):
    group = canvas.get_group(group_id)
    if group is None:
        return

    for box in group.widgets:
        if box is not None:
            box.update_positions()

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def animate_before_join(group_id: int):
    canvas.qobject.groups_to_join.append(group_id)
    group = canvas.get_group(group_id)
    if group is None:
        return

    for widget in group.widgets:
        canvas.scene.add_box_to_animation(
            widget, group.null_pos.x(), group.null_pos.y(), joining=True)

@patchbay_api
def move_group_boxes(
        group_id: int, null_xy: tuple[int, int], in_xy: tuple[int, int],
        out_xy: tuple[int, int], animate=True):
    group = canvas.get_group(group_id)
    if group is None:
        return

    group.null_pos = QPoint(*null_xy)
    group.in_pos = QPoint(*in_xy)
    group.out_pos = QPoint(*out_xy)

    if group.split:
        for port_mode in (PortMode.OUTPUT, PortMode.INPUT):
            box = group.widgets[0]
            xy = out_xy
            pos = group.out_pos

            if port_mode is PortMode.INPUT:
                box = group.widgets[1]
                xy = in_xy
                pos = group.in_pos

            if box is None:
                continue

            box_pos = box.pos()

            if int(box_pos.x()) == xy[0] and int(box_pos.y()) == xy[1]:
                continue

            canvas.scene.add_box_to_animation(
                box, xy[0], xy[1], force_anim=animate)
    else:
        box = group.widgets[0]
        if box is None:
            return

        box_pos = box.pos()
        if int(box_pos.x()) == null_xy[0] and int(box_pos.y()) == null_xy[1]:
            return
        
        canvas.scene.add_box_to_animation(box, null_xy[0], null_xy[1],
                                          force_anim=animate)

@patchbay_api
def wrap_group_box(group_id: int, port_mode: int, yesno: bool,
                   animate=True, prevent_overlap=True):
    group = canvas.get_group(group_id)
    if group is None:
        return

    for box in group.widgets:
        if (box is not None
                and box.get_splitted_mode() == port_mode):
            box.set_wrapped(yesno, animate=animate,
                            prevent_overlap=prevent_overlap)

@patchbay_api
def set_group_layout_mode(group_id: int, port_mode: PortMode,
                          layout_mode: BoxLayoutMode):
    group = canvas.get_group(group_id)
    if group is None:
        _logger.warning(
            "set_group_layout_mode, no group with group_id {group_id}")
        return
    
    group.layout_modes[port_mode] = layout_mode
    
    if canvas.loading_items:
        return

    for box in group.widgets:
        if box is not None:
            box.update_positions()
                        

# ------------------------------------------------------------------------

@patchbay_api
def get_group_pos(group_id, port_mode=PortMode.OUTPUT):
    # Not used now
    group = canvas.get_group(group_id)
    if group is None:
        _logger.error(f"{_logging_str} - unable to find group")
        return QPointF(0, 0)
    
    return group.widgets[1 if (group.split and port_mode is PortMode.INPUT) else 0].pos()

@patchbay_api
def restore_group_positions(data_list):
    # Not used now
    mapping = {}

    for group in canvas.group_list:
        mapping[group.group_name] = group

    for data in data_list:
        name = data['name']
        group = mapping.get(name, None)

        if group is None:
            continue

        assert isinstance(group, GroupObject)

        group.widgets[0].setPos(data['pos1x'], data['pos1y'])

        if group.split and group.widgets[1]:
            group.widgets[1].setPos(data['pos2x'], data['pos2y'])

@patchbay_api
def set_group_pos(group_id, group_pos_x, group_pos_y):
    # Not used now
    set_group_pos_full(group_id, group_pos_x, group_pos_y, group_pos_x, group_pos_y)

@patchbay_api
def set_group_pos_full(group_id, group_pos_x_o, group_pos_y_o,
                       group_pos_x_i, group_pos_y_i):
    # Not used now    
    group = canvas.get_group(group_id)
    if group is None:
        _logger.error(f"set_group_pos_full({group_id}, {group_pos_x_o}, {group_pos_y_o}"
                      f"{group_pos_x_i}, {group_pos_y_i})"
                      " - unable to find group to reposition")
        return

    group.widgets[0].setPos(group_pos_x_o, group_pos_y_o)

    if group.split and group.widgets[1]:
        group.widgets[1].setPos(group_pos_x_i, group_pos_y_i)

    QTimer.singleShot(0, canvas.scene.update)
    

@patchbay_api
def set_group_icon(group_id: int, box_type: BoxType, icon_name: str):
    group = canvas.get_group(group_id)
    if group is None:
        _logger.critical(f"{_logging_str} - unable to find group to change icon")
        return
    
    group.box_type = box_type
    group.icon_name = icon_name

    for widget in group.widgets:
        if widget is not None:
            widget.set_icon(box_type, icon_name)

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def set_group_as_plugin(group_id: int, plugin_id: int,
                        has_ui: bool, has_inline_display: bool):
    group = canvas.get_group(group_id)
    if group is None:
        _logger.critical(f"{_logging_str} - unable to find group to set as plugin")
        return

    group.plugin_id = plugin_id
    group.plugin_ui = has_ui
    group.plugin_inline = has_inline_display
    group.widgets[0].set_as_plugin(plugin_id, has_ui, has_inline_display)

    if group.split and group.widgets[1]:
        group.widgets[1].set_as_plugin(plugin_id, has_ui, has_inline_display)

    canvas.group_plugin_map[plugin_id] = group

# ------------------------------------------------------------------------------------------------------------
@patchbay_api
def add_port(group_id: int, port_id: int, port_name: str,
             port_mode: PortMode, port_type: PortType,
             port_subtype: PortSubType):
    if canvas.get_port(group_id, port_id) is not None:
        _logger.critical(f"{_logging_str} - port already exists")

    group = canvas.get_group(group_id)
    if group is None:
        _logger.critical(f"{_logging_str} - Unable to find parent group")
        return

    n = 0
    if (group.split
            and group.widgets[0].get_splitted_mode() != port_mode
            and group.widgets[1] is not None):
        n = 1

    box_widget = group.widgets[n]

    port = PortObject()
    port.group_id = group_id
    port.port_id = port_id
    port.port_name = port_name
    port.port_mode = port_mode
    port.port_type = port_type
    port.portgrp_id = 0
    port.port_subtype = port_subtype
    port.widget = PortWidget(port, box_widget)
    port.widget.setVisible(not box_widget.is_wrapped())
    canvas.add_port(port)

    canvas.last_z_value += 1
    port.widget.setZValue(canvas.last_z_value)

    if canvas.loading_items:
        return

    box_widget.update_positions()

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def remove_port(group_id: int, port_id: int):
    port = canvas.get_port(group_id, port_id)
    if port is None:
        _logger.critical(f"{_logging_str} - Unable to find port to remove")
        return

    if port.portgrp_id:
        _logger.critical(f"{_logging_str} - Port is in portgroup " 
                            f"{port.portgrp_id}, remove it before !")
        return

    item = port.widget
    box = None
    
    if item is not None:
        box = item.parentItem()
        canvas.scene.removeItem(item)

    del item
    canvas.remove_port(port)

    canvas.qobject.port_removed.emit(group_id, port_id)
    if canvas.loading_items:
        return

    if box is not None:
        box.update_positions()

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def rename_port(group_id: int, port_id: int, new_port_name: str):
    port = canvas.get_port(group_id, port_id)
    if port is None:
        _logger.critical(f"{_logging_str} - Unable to find port to rename")
        return

    if new_port_name != port.port_name:
        port.port_name = new_port_name
        port.widget.set_port_name(new_port_name)

    if canvas.loading_items:
        return

    port.widget.parentItem().update_positions()

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def add_portgroup(group_id: int, portgrp_id: int, port_mode: PortMode,
                  port_type: PortType, port_subtype: PortSubType,
                  port_id_list: list[int]):
    if canvas.get_portgroup(group_id, portgrp_id) is not None:
        _logger.critical(f"{_logging_str} - portgroup already exists")
        return
    
    portgrp = PortgrpObject()
    portgrp.group_id = group_id
    portgrp.portgrp_id = portgrp_id
    portgrp.port_mode = port_mode
    portgrp.port_type = port_type
    portgrp.port_subtype = port_subtype
    portgrp.port_id_list = tuple(port_id_list)
    portgrp.widget = None

    i = 0
    # check that port ids are present and groupable in this group
    for port in canvas.list_ports(group_id=group_id):
        if (port.port_type == port_type
                and port.port_mode == port_mode):
            if port.port_id == port_id_list[i]:
                if port.portgrp_id:
                    _logger.error(
                        f"{_logging_str} - "
                        f"port id {port.port_id} is already in portgroup {port.portgrp_id}")
                    return

                i += 1

                if i == len(port_id_list):
                    # everything seems ok for this portgroup, stop the check
                    break

            elif i > 0:
                _logger.critical(f"{_logging_str} - port ids are not consecutive")
                return
    else:
        _logger.critical(f"{_logging_str} - not enought ports with port_id_list")
        return

    # modify ports impacted by portgroup
    for port in canvas.list_ports(group_id=group_id):
        if (port.port_id in port_id_list):
            port.set_portgroup_id(
                portgrp_id, port_id_list.index(port.port_id), len(port_id_list))

    canvas.add_portgroup(portgrp)
    
    # add portgroup widget and refresh the view
    group = canvas.get_group(group_id)
    if group is None:
        return
    
    for box in group.widgets:
        if box is None:
            continue

        if (not box.is_splitted()
                or box.get_splitted_mode() == port_mode):
            portgrp.widget = box.add_portgroup_from_group(portgrp)

            if not canvas.loading_items:
                box.update_positions()

@patchbay_api
def remove_portgroup(group_id: int, portgrp_id: int):
    box_widget = None

    for portgrp in canvas.list_portgroups(group_id=group_id):
        if portgrp.portgrp_id == portgrp_id:
            # set portgrp_id to the concerned ports
            for port in canvas.list_ports(group_id=group_id):
                if port.portgrp_id == portgrp_id:
                    port.set_portgroup_id(0, 0, 1)

            if portgrp.widget is not None:
                item = portgrp.widget
                box_widget = item.parentItem()
                canvas.scene.removeItem(item)
                del item
                portgrp.widget = None
            break
    else:
        _logger.critical(f"{_logging_str} - Unable to find portgrp to remove")
        return

    canvas.remove_portgroup(portgrp)

    if canvas.loading_items:
        return

    if box_widget is not None:
        box_widget.update_positions()

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def connect_ports(connection_id: int, group_out_id: int, port_out_id: int,
                  group_in_id: int, port_in_id: int):    
    out_port = canvas.get_port(group_out_id, port_out_id)
    in_port = canvas.get_port(group_in_id, port_in_id)
    
    if out_port is None or in_port is None:
        _logger.critical(f"{_logging_str} - unable to find ports to connect")
        return
    
    out_port_wg, in_port_wg = out_port.widget, in_port.widget

    connection = ConnectionObject()
    connection.connection_id = connection_id
    connection.group_in_id = group_in_id
    connection.port_in_id = port_in_id
    connection.group_out_id = group_out_id
    connection.port_out_id = port_out_id
    connection.widget = LineWidget(connection_id, out_port_wg, in_port_wg)

    canvas.scene.addItem(connection.widget)

    canvas.add_connection(connection)
    out_port_wg.parentItem().add_line_to_box(connection.widget)
    in_port_wg.parentItem().add_line_to_box(connection.widget)
    out_port_wg.add_line_to_port(connection.widget)
    in_port_wg.add_line_to_port(connection.widget)

    canvas.qobject.connection_added.emit(connection_id)
    
    if canvas.loading_items:
        return

    QTimer.singleShot(0, canvas.scene.update)

@patchbay_api
def disconnect_ports(connection_id: int):
    connection = canvas.get_connection(connection_id)
    if connection is None:
        _logger.critical(f"{_logging_str} - unable to find connection ports")
        return
    
    tmp_conn = connection.copy_no_widget()
    line = connection.widget
    canvas.remove_connection(connection)
    canvas.qobject.connection_removed.emit(connection_id)

    out_port = canvas.get_port(tmp_conn.group_out_id, tmp_conn.port_out_id)
    in_port = canvas.get_port(tmp_conn.group_in_id, tmp_conn.port_in_id)
    
    if out_port is None or in_port is None:
        canvas.scene.removeItem(line)
        del line
    
        _logger.info(f"{_logging_str} - connection cleaned after its ports")
        return
        
    item1 = out_port.widget
    item2 = in_port.widget
    if item1 is None or item2 is None:
        _logger.critical(f"{_logging_str} - port has no widget")
        return

    item1.parentItem().remove_line_from_box(connection)
    item2.parentItem().remove_line_from_box(connection)
    item1.remove_line_from_port(connection)
    item2.remove_line_from_port(connection)

    canvas.scene.removeItem(line)
    del line

    if canvas.loading_items:
        return

    QTimer.singleShot(0, canvas.scene.update)

# ----------------------------------------------------------------------------

@patchbay_api
def get_theme() -> str:
    return canvas.theme_manager.get_theme()

@patchbay_api
def list_themes() -> list[dict]:
    return canvas.theme_manager.list_themes()

@patchbay_api
def change_theme(theme_name='') -> bool:
    return canvas.theme_manager.set_theme(theme_name)

@patchbay_api
def copy_and_load_current_theme(new_theme_name: str) -> int:
    return canvas.theme_manager.copy_and_load_current_theme(new_theme_name)

# ----------------------------------------------------------------------------
@patchbay_api
def redraw_plugin_group(plugin_id: int):
    group = canvas.group_plugin_map.get(plugin_id, None)

    if group is None:
        _logger.critical(f"{_logging_str} - unable to find group")
        return

    assert isinstance(group, GroupObject)

    group.widgets[0].redraw_inline_display()

    if group.split and group.widgets[1]:
        group.widgets[1].redraw_inline_display()

@patchbay_api
def handle_plugin_removed(plugin_id: int):
    group = canvas.group_plugin_map.pop(plugin_id, None)

    if group is not None:
        assert isinstance(group, GroupObject)
        group.plugin_id = -1
        group.plugin_ui = False
        group.plugin_inline = False
        group.widgets[0].remove_as_plugin()

        if group.split and group.widgets[1]:
            group.widgets[1].remove_as_plugin()

    for group in canvas.group_list:
        if group.plugin_id < plugin_id or group.plugin_id > MAX_PLUGIN_ID_ALLOWED:
            continue

        group.plugin_id -= 1
        group.widgets[0]._plugin_id -= 1

        if group.split and group.widgets[1]:
            group.widgets[1]._plugin_id -= 1

        canvas.group_plugin_map[plugin_id] = group

@patchbay_api
def handle_all_plugins_removed():
    canvas.group_plugin_map = {}

    for group in canvas.group_list:
        if group.plugin_id < 0:
            continue
        if group.plugin_id > MAX_PLUGIN_ID_ALLOWED:
            continue

        group.plugin_id = -1
        group.plugin_ui = False
        group.plugin_inline = False
        group.widgets[0].remove_as_plugin()

        if group.split and group.widgets[1]:
            group.widgets[1].remove_as_plugin()

@patchbay_api
def set_auto_select_items(yesno: bool):
    options.auto_select_items = yesno
    
    for box in canvas.list_boxes():
        box.setAcceptHoverEvents(yesno)
    
    for portgrp in canvas.list_portgroups():
        if portgrp.widget is not None:
            portgrp.widget.setAcceptHoverEvents(yesno)
            
    for port in canvas.list_ports():
        if port.widget is not None:
            port.widget.setAcceptHoverEvents(yesno)

@patchbay_api
def set_elastic(yesno: bool):
    canvas.scene.set_elastic(yesno)

@patchbay_api
def set_prevent_overlap(yesno: bool):
    options.prevent_overlap = yesno
    if yesno:
        redraw_all_groups()

@patchbay_api
def set_borders_navigation(yesno: bool):
    options.borders_navigation = yesno

@patchbay_api
def set_max_port_width(width: int):
    options.max_port_width = width
    redraw_all_groups()

@patchbay_api
def set_default_zoom(default_zoom: int):
    options.default_zoom = default_zoom

@patchbay_api
def semi_hide_group(group_id: int, yesno: bool):
    group = canvas.get_group(group_id)
    if group is None:
        return

    for widget in group.widgets:
        if widget is not None:
            widget.semi_hide(yesno)

@patchbay_api
def semi_hide_connection(connection_id: int, yesno: bool):
    connection = canvas.get_connection(connection_id)
    if connection and connection.widget is not None:
        connection.widget.semi_hide(yesno)

@patchbay_api
def set_group_in_front(group_id: int):
    canvas.last_z_value += 1
    group = canvas.get_group(group_id)
    if group is None:
        return
    
    for widget in group.widgets:
        if widget is not None:
            widget.setZValue(canvas.last_z_value)

@patchbay_api
def set_connection_in_front(connection_id: int):
    canvas.last_z_value += 1
    
    connection = canvas.get_connection(connection_id)
    if connection and connection.widget is not None:
        connection.widget.setZValue(canvas.last_z_value)

@patchbay_api
def select_port(group_id: int, port_id: int):
    port = canvas.get_port(group_id, port_id)
    if port is None:
        return
    
    if port.widget is None:
        return
    
    box_widget = port.widget.parentItem()
    canvas.scene.clearSelection()

    if box_widget.is_wrapped():
        canvas.scene.center_view_on(box_widget)
        box_widget.setSelected(True)
    else:
        canvas.scene.center_view_on(port.widget)
        port.widget.setSelected(True)

@patchbay_api
def select_filtered_group_box(group_id: int, n_select=1):
    group = canvas.get_group(group_id)
    if group is None:
        return
    
    n_widget = 1

    for widget in group.widgets:
        if widget is not None and widget.isVisible():
            if n_select == n_widget:
                canvas.scene.clearSelection()
                widget.setSelected(True)
                canvas.scene.center_view_on(widget)
                break

            n_widget += 1

@patchbay_api
def get_number_of_boxes(group_id: int) -> int:
    group = canvas.get_group(group_id)
    if group is None:
        return 0
    
    n = 0
    for widget in group.widgets:
        if widget is not None and widget.isVisible():
            n += 1
    
    return n

@patchbay_api    
def set_semi_hide_opacity(opacity: float):
    options.semi_hide_opacity = opacity

    for box in canvas.list_boxes():
        box.update_opacity()
                
    for conn in canvas.list_connections():
        if conn.widget is not None:
            conn.widget.update_line_gradient()

@patchbay_api
def set_optional_gui_state(group_id: int, visible: bool):
    group = canvas.get_group(group_id)
    if group is None:
        return

    group.handle_client_gui = True
    group.gui_visible = visible

    for widget in group.widgets:
        if widget is not None:
            widget.set_optional_gui_state(visible)
    
    if not canvas.loading_items:
        canvas.scene.update()

@patchbay_api
def zoom_reset():
    if canvas.scene is None:
        return

    canvas.scene.zoom_reset()
    
@patchbay_api
def zoom_fit():
    if canvas.scene is None:
        return
    
    canvas.scene.zoom_fit()

@patchbay_api
def save_cache():
    canvas.theme.save_cache()

@patchbay_api
def set_grouped_box_layout_ratio(value: float):
    options.box_grouped_auto_layout_ratio = max(min(2.0, value), 0.0)
    redraw_all_groups()

@patchbay_api
def set_options(new_options: CanvasOptionsObject):
    if not canvas.initiated:
        options.__dict__ = new_options.__dict__.copy()

@patchbay_api
def set_features(new_features: CanvasFeaturesObject):
    if not canvas.initiated:
        features.__dict__ = new_features.__dict__.copy()



from queue import Queue
import time
from enum import IntEnum
from typing import Iterator, Optional


class PortMode(IntEnum):
    NULL = 0
    OUTPUT = 1
    INPUT = 2


# It is here if we want to improve the saved file
# with the type of the port.
# At this stage, we only care about the port name.
class PortType(IntEnum):
    NULL = 0
    AUDIO = 1
    MIDI = 2


class JackPort:
    # is_new is used to prevent reconnections
    # when a disconnection has not been saved and one new port append.
    id = 0
    name = ''
    mode = PortMode.NULL
    type = PortType.NULL
    is_new = False
    
    
class ProtoEngine:
    def init(self) -> bool:
        return True

    def fill_ports_and_connections(
            self, port_list: dict[PortMode, list[JackPort]],
            connection_list: list[tuple[str, str]]):
        ...
    def connect_ports(self, port_out: str, port_in: str):
        ...
    def disconnect_ports(self, port_out: str, port_in: str):
        ...
    def quit(self):
        ...


class Event(IntEnum):
    CLIENT_ADDED = 1
    CLIENT_REMOVED = 2
    PORT_ADDED = 3
    PORT_REMOVED = 4
    PORT_RENAMED = 5
    CONNECTION_ADDED = 6
    CONNECTION_REMOVED = 7
    JACK_STOPPED = 8


class MonitorStates(IntEnum):
    NEVER_DONE = 0
    UPDATING = 1
    DONE = 2


class Timer:
    _last_ask = 0.0
    _duration: float
    
    def __init__(self, duration: float):
        self._duration = duration
        
    def start(self):
        self._last_ask = time.time()
        
    def elapsed(self) -> bool:
        if not self._last_ask:
            return False
        
        elapsed = time.time() - self._last_ask >= self._duration
        if elapsed:
            self._last_ask = 0.0
        return elapsed


class EventHandler:
    _event_queue = Queue()
    
    @classmethod
    def add_event(cls, event: Event, *args: tuple):
        cls._event_queue.put((event, args))

    @classmethod
    def new_events(cls) -> Iterator[tuple[Event, tuple]]:
        while cls._event_queue.qsize():
            yield cls._event_queue.get()


class Glob:
    file_path = ''
    is_dirty = False
    dirty_state_sent = False
    pending_connection = False
    open_done_once = False
    allow_disconnections = False
    terminate = False
    monitor_states_done = MonitorStates.NEVER_DONE
    client_changing_id: Optional[tuple[str, str]] = None


def b2str(src_bytes: bytes) -> str:
    ''' decodes bytes to string '''
    return str(src_bytes, encoding="utf-8")

def debug_conn_str(conn: tuple[str, str]):
    return f"connection from '{conn[0]}' to '{conn[1]}'"





from typing import TYPE_CHECKING
from PyQt5.QtWidgets import (
    QToolBar, QLabel, QMenu,
    QApplication, QAction)
from PyQt5.QtGui import QMouseEvent, QIcon
from PyQt5.QtCore import pyqtSignal, Qt, QPoint, QSize


from .base_elements import ToolDisplayed
from .patchbay_manager import PatchbayManager
from .tools_widgets import PatchbayToolsWidget

_translate = QApplication.translate


class PatchbayToolBar(QToolBar):    
    def __init__(self, parent):
        super().__init__(parent)
        self.setContextMenuPolicy(Qt.PreventContextMenu)

        self._displayed_widgets = (
            ToolDisplayed.ZOOM_SLIDER
            | ToolDisplayed.TRANSPORT_PLAY_STOP
            | ToolDisplayed.BUFFER_SIZE
            | ToolDisplayed.SAMPLERATE
            | ToolDisplayed.XRUNS
            | ToolDisplayed.DSP_LOAD)
        
        self._transport_widget = None
        self._patchbay_mng : PatchbayManager = None
    
    def set_patchbay_manager(self, patchbay_manager: PatchbayManager):
        self._patchbay_mng = patchbay_manager
        patchbay_manager.change_tools_displayed(self._displayed_widgets)
    
    def _change_visibility(self):
        if self._patchbay_mng is not None:
            self._patchbay_mng.change_tools_displayed(self._displayed_widgets)

    def set_default_displayed_widgets(self, displayed_widgets: ToolDisplayed):
        self._displayed_widgets = displayed_widgets
        self._change_visibility()
        
    def get_displayed_widgets(self) -> ToolDisplayed:
        return self._displayed_widgets

    def _make_context_actions(self) -> dict[ToolDisplayed, QAction]:
        return {
            ToolDisplayed.PORT_TYPES_VIEW:
                QAction(QIcon.fromTheme('view-filter'),
                        _translate('tool_bar', 'Type filter')),
            ToolDisplayed.ZOOM_SLIDER:
                QAction(QIcon.fromTheme('zoom-select'),
                        _translate('tool_bar', 'Zoom slider')),
            ToolDisplayed.TRANSPORT_CLOCK:
                QAction(QIcon.fromTheme('clock'),
                        _translate('tool_bar', 'Transport clock')),
            ToolDisplayed.TRANSPORT_PLAY_STOP:
                QAction(QIcon.fromTheme('media-playback-pause'),
                        _translate('tool_bar', 'Transport Play/Stop')),
            ToolDisplayed.TRANSPORT_TEMPO:
                QAction(QIcon.fromTheme('folder-music-symbolic'),
                        _translate('tool_bar', 'Transport Tempo')),
            ToolDisplayed.BUFFER_SIZE:
                QAction(QIcon.fromTheme('settings-configure'),
                        _translate('tool_bar', 'Buffer size')),
            ToolDisplayed.SAMPLERATE:
                QAction(QIcon.fromTheme('filename-sample-rate'),
                        _translate('tool_bar', 'Sample rate')),
            ToolDisplayed.LATENCY:
                QAction(QIcon.fromTheme('chronometer-lap'),
                        _translate('tool_bar', 'Latency')),
            ToolDisplayed.XRUNS:
                QAction(QIcon.fromTheme('data-error'),
                        _translate('tool_bar', 'Xruns')),
            ToolDisplayed.DSP_LOAD:
                QAction(QIcon.fromTheme('histogram-symbolic'),
                        _translate('tool_bar', 'DSP Load'))
        }

    def _make_context_menu(self, context_actions: dict[ToolDisplayed, QAction]) -> QMenu:
        menu = QMenu()
        menu.addSection(_translate('tool_bar', 'Displayed tools'))
        
        for key, act in context_actions.items():
            act.setCheckable(True)
            act.setChecked(bool(self._displayed_widgets & key))
            if self._patchbay_mng is not None:
                if not self._patchbay_mng.server_is_started:
                    if key in (ToolDisplayed.PORT_TYPES_VIEW, ToolDisplayed.ZOOM_SLIDER):
                        act.setEnabled(self._patchbay_mng.alsa_midi_enabled)
                    else:
                        act.setEnabled(False)
            menu.addAction(act)
            
        return menu

    def mousePressEvent(self, event: QMouseEvent) -> None:
        child_widget = self.childAt(event.pos())
        super().mousePressEvent(event)

        if (event.button() != Qt.RightButton
                or not isinstance(child_widget, (QLabel, PatchbayToolsWidget))):
            return

        context_actions = self._make_context_actions()
        menu = self._make_context_menu(context_actions)
        
        # execute the menu, exit if no action
        point = event.screenPos().toPoint()
        point.setY(self.mapToGlobal(QPoint(0, self.height())).y())
        selected_act = menu.exec(point)
        if selected_act is None:
            return

        for key, act in context_actions.items():
            if act is selected_act:
                if act.isChecked():
                    self._displayed_widgets |= key
                else:
                    self._displayed_widgets &= ~key

        self._change_visibility()



from dataclasses import dataclass
import time
from typing import TYPE_CHECKING, Iterator, Optional
from threading import Thread
    
from pyalsa import alsaseq
from pyalsa.alsaseq import (
    SEQ_USER_CLIENT,
    SEQ_PORT_CAP_NO_EXPORT,
    SEQ_PORT_CAP_READ,
    SEQ_PORT_CAP_SUBS_READ,
    SEQ_PORT_CAP_WRITE,
    SEQ_PORT_CAP_SUBS_WRITE,
    SEQ_PORT_TYPE_APPLICATION,
    SEQ_CLIENT_SYSTEM,
    SEQ_PORT_SYSTEM_ANNOUNCE,
    SEQ_EVENT_CLIENT_START,
    SEQ_EVENT_CLIENT_EXIT,
    SEQ_EVENT_PORT_START,
    SEQ_EVENT_PORT_EXIT,
    SEQ_EVENT_PORT_SUBSCRIBED,
    SEQ_EVENT_PORT_UNSUBSCRIBED,
    SequencerError
)
from port_data import PortData

if TYPE_CHECKING:
    from ray_jackpatch_to_osc import MainObject
    from osc_server import OscJackPatch

PORT_IS_INPUT = 0x1
PORT_IS_OUTPUT = 0x2
PORT_IS_PHYSICAL = 0x4

PORT_TYPE_MIDI_ALSA = 0x04

_PORT_READS = SEQ_PORT_CAP_READ | SEQ_PORT_CAP_SUBS_READ
_PORT_WRITES = SEQ_PORT_CAP_WRITE | SEQ_PORT_CAP_SUBS_WRITE


@dataclass
class AlsaPort:
    name: str
    id: int
    caps: int
    physical: bool
    
    def pb_name(self, mode_str: str, client: 'AlsaClient'):
        return f":ALSA_{mode_str}:{client.id}:{self.id}:{client.name}:{self.name}"

@dataclass
class AlsaConn:
    source_client_id: int
    source_port_id: int
    dest_client_id: int
    dest_port_id: int
    
    def as_port_names(self, clients: dict[int, 'AlsaClient']) -> Optional[tuple[str, str]]:
        src_client = clients.get(self.source_client_id)
        dest_client = clients.get(self.dest_client_id)
        
        if src_client is None or dest_client is None:
            return None
        
        src_port = src_client.ports.get(self.source_port_id)
        dest_port = dest_client.ports.get(self.dest_port_id)
        
        if src_port is None or dest_port is None:
            return None
        
        return (src_port.pb_name('OUT', src_client),
                dest_port.pb_name('IN', dest_client))


class AlsaClient:
    def __init__(self, alsa_mng: 'AlsaManager', name: str, id: int):
        self.alsa_mng = alsa_mng
        self.name = name
        self.id = id
        self.ports = dict[int, AlsaPort]()
            
    def __repr__(self) -> str:
        return f"AlsaClient({self.name}, {self.id})"
    
    def add_port(self, port_id: int):
        try:
            port_info = self.alsa_mng.seq.get_port_info(port_id, self.id)
            caps = port_info['capability']
        except:
            return

        if caps & SEQ_PORT_CAP_NO_EXPORT:
            return
        
        if not(caps & _PORT_READS == _PORT_READS
               or caps & _PORT_WRITES == _PORT_WRITES):
            return

        physical = not bool(port_info['type'] & SEQ_PORT_TYPE_APPLICATION)
        self.ports[port_id] = AlsaPort(port_info['name'], port_id, caps, physical)


class AlsaManager:
    def __init__(self, jack_mng: 'MainObject'):
        self._jack_mng = jack_mng
        self._osc_server = self._jack_mng.osc_server
        self.seq = alsaseq.Sequencer(clientname='raysession')

        self._all_alsa_connections = list[AlsaConn]()
        self._connections = list[AlsaConn]()
        self._clients = dict[int, AlsaClient]()
        self._clients_names = dict[int, str]()

        self._stopping = False
        self._event_thread = Thread(target=self.read_events)

        port_caps = (SEQ_PORT_CAP_WRITE
                     | SEQ_PORT_CAP_SUBS_WRITE
                     | SEQ_PORT_CAP_NO_EXPORT)
        input_id = self.seq.create_simple_port(
            name="raysession_port",
            type=SEQ_PORT_TYPE_APPLICATION,
            caps=port_caps)

        self.seq.connect_ports(
            (SEQ_CLIENT_SYSTEM, SEQ_PORT_SYSTEM_ANNOUNCE),
            (self.seq.client_id, input_id))

    def get_the_graph(self):
        if self.seq is None:
            return
        
        clients = self.seq.connection_list()
        
        self._clients_names.clear()
        self._clients.clear()
        self._all_alsa_connections.clear()

        for client in clients:
            client_name, client_id, port_list = client
            self._clients_names[client_id] = client_name
            self._clients[client_id] = AlsaClient(
                self, client_name, client_id)
            for port_name, port_id, connection_list in port_list:
                self._clients[client_id].add_port(port_id)
      
                connections = connection_list[0]
                for connection in connections:
                    conn_client_id, conn_port_id = connection[:2]                    
                    self._all_alsa_connections.append(
                        AlsaConn(client_id, port_id,
                                 conn_client_id, conn_port_id))

    def add_port_to_patchbay(self, client: AlsaClient, port: AlsaPort):
        port_flags = 0
        if port.physical:
            port_flags = PORT_IS_PHYSICAL
        
        if port.caps & _PORT_READS == _PORT_READS:
            self._osc_server.port_added(
                port.pb_name('OUT', client),
                PORT_TYPE_MIDI_ALSA,
                port_flags | PORT_IS_OUTPUT,
                client.id * 0x10000 + port.id)

        if port.caps & _PORT_WRITES == _PORT_WRITES:
            self._osc_server.port_added(
                port.pb_name('IN', client),
                PORT_TYPE_MIDI_ALSA,
                port_flags | PORT_IS_INPUT,
                client.id * 0x10000 + port.id)

    def remove_port_from_patchbay(self, client: AlsaClient, port: AlsaPort):
        if port.caps & _PORT_READS == _PORT_READS:
            self._osc_server.port_removed(
                port.pb_name('OUT', client))
        if port.caps & _PORT_WRITES == _PORT_WRITES:
            self._osc_server.port_removed(
                port.pb_name('IN', client))

    def add_all_ports(self):        
        if self._event_thread.is_alive():
            self.stop_events_loop()

        self.get_the_graph()

        for client in self._clients.values():
            if client.name == 'System':
                continue

            for port in client.ports.values():
                self.add_port_to_patchbay(client, port)
                
        for conn in self._all_alsa_connections:
            source_client = self._clients.get(conn.source_client_id)
            dest_client = self._clients.get(conn.dest_client_id)
            if source_client is None or dest_client is None:
                continue
            
            source_port = source_client.ports.get(conn.source_port_id)
            dest_port = dest_client.ports.get(conn.dest_port_id)
            
            if source_port is None or dest_port is None:
                continue
            
            self._connections.append(conn)
            
            self._osc_server.connection_added(
                (source_port.pb_name('OUT', source_client),
                 dest_port.pb_name('IN', dest_client))
            )
    
        self._event_thread.start()
    
    def parse_ports_and_flags(self) -> Iterator[PortData]:
        for client_id, client in self._clients.items():
            if client.name == 'System':
                continue
            
            for port_id, port in client.ports.items():
                port_flags = 0
                if port.physical:
                    port_flags = PORT_IS_PHYSICAL

                if port.caps & _PORT_READS == _PORT_READS:
                    yield PortData(
                        port.pb_name('OUT', client),
                        PORT_TYPE_MIDI_ALSA,
                        port_flags | PORT_IS_OUTPUT,
                        client.id * 0x10000 + port.id)
                
                if port.caps & _PORT_WRITES == _PORT_WRITES:
                    yield PortData(
                        port.pb_name('IN', client),
                        PORT_TYPE_MIDI_ALSA,
                        port_flags | PORT_IS_INPUT,
                        client.id * 0x10000 + port.id)
    
    def parse_connections(self) -> Iterator[tuple[str, str]]:
        for conn in self._connections:
            src_client = self._clients.get(conn.source_client_id)
            dest_client = self._clients.get(conn.dest_client_id)
            if src_client is None or dest_client is None:
                continue
            
            src_port = src_client.ports.get(conn.source_port_id)
            dest_port = dest_client.ports.get(conn.dest_port_id)
            
            if src_port is None or dest_port is None:
                continue
            
            yield (src_port.pb_name('OUT', src_client),
                   dest_port.pb_name('IN', dest_client))
    
    def connect_ports(self, port_out_name: str, port_in_name: str,
                      disconnect=False):
        try:
            _, alsa_key, src_client_id, src_port_id, *rest = port_out_name.split(':')
            _, alsa_key, dest_client_id, dest_port_id, *rest = port_in_name.split(':')            
            src_client_id, src_port_id = int(src_client_id), int(src_port_id)
            dest_client_id, dest_port_id = int(dest_client_id), int(dest_port_id)
        except:
            # TODO log
            return
        
        try:
            if disconnect:
                self.seq.disconnect_ports(
                    (src_client_id, src_port_id),
                    (dest_client_id, dest_port_id))
            else: 
                self.seq.connect_ports(
                    (src_client_id, src_port_id),
                    (dest_client_id, dest_port_id),
                    0, 0, 0, 0)
        except Exception as e:
            # TODO log something
            pass
        
    def read_events(self):
        while True:
            if self._stopping:
                break

            event_list = self.seq.receive_events(timeout=128, maxevents=1)

            for event in event_list:
                data = event.get_data()

                if event.type == SEQ_EVENT_CLIENT_START:
                    try:
                        client_id = data['addr.client']
                        client_info = self.seq.get_client_info(client_id)
                    except:
                        continue

                    n_tries = 0
                    client_outed = False

                    # Sometimes client name is not ready
                    while client_info['name'] == f'Client-{client_id}':
                        time.sleep(0.010)
                        try:
                            client_info = self.seq.get_client_info(client_id)
                        except:
                            client_outed = True
                            break
                        
                        n_tries += 1
                        if n_tries >= 5:
                            break
                    
                    if client_outed:
                        continue

                    self._clients[client_id] = AlsaClient(
                        self, client_info['name'], client_id)

                elif event.type == SEQ_EVENT_CLIENT_EXIT:
                    client_id = data['addr.client']
                    client = self._clients.get(client_id)
                    if client is not None:
                        for port in client.ports.values():
                            self.remove_port_from_patchbay(client, port)

                        del self._clients[client_id]
                    
                elif event.type == SEQ_EVENT_PORT_START:
                    client_id, port_id = data['addr.client'], data['addr.port']
                    client = self._clients.get(client_id)
                    if client is None:
                        continue
                    
                    client.add_port(port_id)
                    port = client.ports.get(port_id)
                    if port is None:
                        continue
                    
                    self.add_port_to_patchbay(client, port)
                    
                elif event.type == SEQ_EVENT_PORT_EXIT:
                    client_id, port_id = data['addr.client'], data['addr.port']
                    client = self._clients.get(client_id)
                    if client is None:
                        continue

                    port = client.ports.get(port_id)
                    if port is None:
                        continue
                    
                    to_rm_conns = list[AlsaConn]()
                    
                    for conn in self._connections:
                        if (client_id, port_id) in ((conn.source_client_id, conn.source_port_id),
                                                    (conn.dest_client_id, conn.dest_port_id)):
                            to_rm_conns.append(conn)
                            
                    for conn in to_rm_conns:
                        port_names = conn.as_port_names(self._clients)
                        if port_names is not None:
                            self._osc_server.connection_removed(port_names)
                    
                    self.remove_port_from_patchbay(client, port)
                    
                elif event.type == SEQ_EVENT_PORT_SUBSCRIBED:
                    sender_client = self._clients.get(data['connect.sender.client'])
                    dest_client = self._clients.get(data['connect.dest.client'])
                    if sender_client is None or dest_client is None:
                        continue
                    
                    sender_port = sender_client.ports.get(data['connect.sender.port'])
                    dest_port = dest_client.ports.get(data['connect.dest.port'])
                    
                    if sender_port is None or dest_port is None:
                        continue

                    self._connections.append(
                        AlsaConn(sender_client.id, sender_port.id,
                                 dest_client.id, dest_port.id))

                    self._osc_server.connection_added(
                        (sender_port.pb_name('OUT', sender_client),
                         dest_port.pb_name('IN', dest_client)))

                elif event.type == SEQ_EVENT_PORT_UNSUBSCRIBED:
                    sender_client = self._clients.get(data['connect.sender.client'])
                    dest_client = self._clients.get(data['connect.dest.client'])
                    if sender_client is None or dest_client is None:
                        continue

                    sender_port = sender_client.ports.get(data['connect.sender.port'])
                    dest_port = dest_client.ports.get(data['connect.dest.port'])

                    if sender_port is None or dest_port is None:
                        continue

                    for conn in self._connections:
                        if (conn.source_client_id == sender_client.id
                                and conn.source_port_id == sender_port.id
                                and conn.dest_client_id == dest_client.id
                                and conn.dest_port_id == dest_port.id):
                            self._connections.remove(conn)
                            break 

                    self._osc_server.connection_removed(
                        (sender_port.pb_name('OUT', sender_client),
                         dest_port.pb_name('IN', dest_client))
                    )
                
    def stop_events_loop(self):
        if not self._event_thread.is_alive():
            return

        self._stopping = True
        self._event_thread.join()
        
        for client in self._clients.values():
            for port in client.ports.values():
                self.remove_port_from_patchbay(client, port)
        
        del self._event_thread
        self._stopping = False
        self._event_thread = Thread(target=self.read_events)
    
    def exit(self):
        self.seq.exit()
    



import sys
from PyQt5.QtWidgets import QApplication

import ray
from client_properties_dialog import NsmClientPropertiesDialog
from patchbay.base_elements import TransportPosition
from daemon_manager import DaemonManager
from gui_client import Client, TrashedClient
from gui_signaler import Signaler
from gui_server_thread import GuiServerThread
from gui_tools import CommandLineArgs, RS, error_text
from main_window import MainWindow
from nsm_child import NsmChild, NsmChildOutside
from ray_patchbay_manager import RayPatchbayManager


class Session:
    def __init__(self):
        self.client_list = list[Client]()
        self.trashed_clients = list[TrashedClient]()
        self.favorite_list = list[ray.Favorite]()
        self.recent_sessions = list[str]()
        self.name = ''
        self.path = ''
        self.notes = ''
        self.terminal_command = ''
        self.server_status = ray.ServerStatus.OFF

        self.is_renameable = True

        self.signaler = Signaler()
        self.patchbay_manager = RayPatchbayManager(self)

        server = GuiServerThread.instance()
        server.start()

        RS.set_signaler(self.signaler)

        self.daemon_manager = DaemonManager(self)
        if CommandLineArgs.daemon_url:
            self.daemon_manager.set_osc_address(CommandLineArgs.daemon_url)
        elif CommandLineArgs.daemon_port:
            self.daemon_manager.set_osc_address(CommandLineArgs.daemon_port)
        elif not CommandLineArgs.out_daemon:
            self.daemon_manager.set_new_osc_address()

        # build nsm_child if NSM_URL in env
        self.nsm_child = None

        if CommandLineArgs.under_nsm:
            if CommandLineArgs.out_daemon:
                self.nsm_child = NsmChildOutside(self)
                self.daemon_manager.set_external()
            else:
                self.nsm_child = NsmChild(self)

        # build and show Main UI
        self.main_win = MainWindow(self)
        self.daemon_manager.finish_init()
        self.patchbay_manager.finish_init()
        server.finish_init(self)
        self.main_win.show()

        # display donations dialog under breizh conditions
        if not RS.is_hidden(RS.HD_Donations):
            coreff_counter = RS.settings.value('coreff_counter', 0, type=int)
            coreff_counter += 1
            RS.settings.setValue('coreff_counter', coreff_counter)

            if coreff_counter % 44 == 29:
                self.main_win.donate(True)

    def quit(self):
        self.patchbay_manager.clear_all()
        self.main_win.hide()
        del self.main_win

    def is_running(self)->bool:
        return bool(self.server_status != ray.ServerStatus.OFF)

    def update_server_status(self, server_status: int):
        self.server_status = server_status

    def _set_name(self, session_name: str):
        self.name = session_name

    def _set_path(self, session_path: str):
        self.path = session_path

    def get_short_path(self):
        if self.path.startswith(CommandLineArgs.session_root):
            return self.path.replace(
                '%s/' % CommandLineArgs.session_root, '', 1)

        return self.path

    def get_client(self, client_id: str)->Client:
        for client in self.client_list:
            if client.client_id == client_id:
                return client

        if CommandLineArgs.debug:
            sys.stderr.write("gui_session does not contains client %s\n"
                             % client_id)
        return None

    def add_favorite(self, template_name: str, icon_name: str,
                     factory: bool, display_name: str):
        server = GuiServerThread.instance()
        if server:
            server.to_daemon('/ray/favorites/add', template_name,
                             icon_name, int(factory), display_name)

    def remove_favorite(self, template_name: str, factory: bool):
        for favorite in self.favorite_list:
            if favorite.name == template_name and favorite.factory == factory:
                break

        server = GuiServerThread.instance()
        if server:
            server.to_daemon('/ray/favorites/remove', template_name, int(factory))

    def is_favorite(self, template_name: str, factory: bool):
        for favorite in self.favorite_list:
            if favorite.name == template_name and favorite.factory == factory:
                return True
        return False

    def set_daemon_options(self, options):
        self.main_win.set_daemon_options(options)
        for client in self.client_list:
            client.widget.set_daemon_options(options)


class SignaledSession(Session):
    def __init__(self):
        Session.__init__(self)
        self.signaler.osc_receive.connect(self._osc_receive)
        self.daemon_manager.start()
        
        self.preview_notes = ''
        self.preview_client_list = list[ray.ClientData]()
        self.preview_started_clients = set()
        self.preview_snapshots = list[str]()
        self.preview_size = -1

    def _osc_receive(self, path: str, args: tuple):
        func_path = path
        func_name = func_path.replace('/', '_')

        if func_name in self.__dir__():
            function = self.__getattribute__(func_name)
            function(path, args)

    def _reply(self, path, args):
        if len(args) == 2:
            if args[0] == '/ray/session/add_executable':
                client_id: str = args[1]

                for client in self.client_list:
                    if (client.client_id == client_id
                            and client.protocol == ray.Protocol.RAY_HACK):
                        client.show_properties_dialog(second_tab=True)
                        break


    def _error(self, path, args):
        err_path, err_code, err_message = args

        # don't shows a window error if error is OK
        # or related to an abort made by user
        if err_code in (ray.Err.OK, ray.Err.ABORT_ORDERED,
                        ray.Err.COPY_ABORTED):
            return

        message = error_text(err_code)
        if message:
            err_message = message

        self.main_win.error_message(err_message)

    def _minor_error(self, path, args):
        err_path, err_code, err_message = args

        # don't shows a window error if error is OK
        # or if it comes from just an unknown (and untreated) message
        if err_code in (ray.Err.OK, ray.Err.UNKNOWN_MESSAGE):
            return

        self.main_win.error_message(err_message)


    def _ray_gui_server_disannounce(self, path, args):
        QApplication.quit()

    def _ray_gui_server_nsm_locked(self, path, args):
        nsm_locked = bool(args[0])
        self.main_win.set_nsm_locked(nsm_locked)

    def _ray_gui_server_message(self, path, args):
        message = args[0]
        self.main_win.print_message(message)

    def _ray_gui_server_terminal_command(self, path, args):
        self.terminal_command = args[0]
        if self.main_win.preferences_dialog is not None:
            self.main_win.preferences_dialog.set_terminal_command(
                self.terminal_command)

    def _ray_gui_server_options(self, path, args):
        options = args[0]
        self.set_daemon_options(options)

    def _ray_gui_server_recent_sessions(self, path, args):
        self.recent_sessions: list[str] = args
        self.main_win.update_recent_sessions_menu()

    def _ray_gui_session_name(self, path, args):
        sname, spath = args
        self._set_name(sname)
        self._set_path(spath)
        self.main_win.rename_session(sname, spath)

    def _ray_gui_session_is_nsm(self, path, args):
        self.main_win.opening_nsm_session()

    def _ray_gui_session_renameable(self, path, args):
        self.is_renameable = bool(args[0])

        bool_set_edit = bool(self.is_renameable
                             and self.server_status == ray.ServerStatus.READY
                             and not CommandLineArgs.out_daemon)

        self.main_win.set_session_name_editable(bool_set_edit)

    def _ray_gui_session_notes(self, path, args):
        self.notes = args[0]
        if self.main_win.notes_dialog is not None:
            self.main_win.notes_dialog.notes_updated()
        
    def _ray_gui_session_notes_shown(self, path, args):
        self.main_win.edit_notes()

    def _ray_gui_session_notes_hidden(self, path, args):
        self.main_win.edit_notes(close=True)

    def _ray_gui_session_sort_clients(self, path, args):
        new_client_list = list[Client]()
        for client_id in args:
            client = self.get_client(client_id)

            if not client:
                return

            new_client_list.append(client)

        if args == [c.client_id for c in self.client_list]:
            # no change between existing and new order
            return

        self.client_list.clear()
        self.main_win.re_create_list_widget()

        self.client_list = new_client_list
        for client in self.client_list:
            client.re_create_widget()
            client.widget.update_status(client.status)

    def _ray_gui_client_new(self, path, args):
        client = Client(self, *args[:2])
        client.update_properties(*args)
        self.client_list.append(client)

    def _ray_gui_client_update(self, path, args):
        client_id = args[0]
        client = self.get_client(client_id)
        if client:
            client.update_properties(*args)

    def _ray_gui_client_ray_hack_update(self, path, args: list):
        client_id = args.pop(0)
        client = self.get_client(client_id)
        if client and client.protocol == ray.Protocol.RAY_HACK:
            client.update_ray_hack(*args)

    def _ray_gui_client_ray_net_update(self, path, args: list):
        client_id = args.pop(0)
        client = self.get_client(client_id)
        if client and client.protocol == ray.Protocol.RAY_NET:
            client.update_ray_net(*args)

    def  _ray_gui_client_switch(self, path, args):
        old_id, new_id = args
        for client in self.client_list:
            if client.client_id == old_id:
                client.client_id = new_id
                break

    def _ray_gui_client_status(self, path, args):
        client_id, status = args
        client = self.get_client(client_id)
        if client:
            client.set_status(status)

            if status == ray.ClientStatus.REMOVED:
                self.main_win.remove_client(client_id)
                client.close_properties_dialog()
                self.client_list.remove(client)
                del client

        self.main_win.client_status_changed(client_id, status)

    def _ray_gui_client_progress(self, path, args):
        client_id, progress = args

        client = self.get_client(client_id)
        if client:
            client.set_progress(progress)

    def _ray_gui_client_dirty(self, path, args):
        client_id, int_dirty = args
        client = self.get_client(client_id)
        if client:
            client.set_dirty_state(bool(int_dirty))

    def _ray_gui_client_gui_visible(self, path, args):
        client_id, int_state = args
        client = self.get_client(client_id)
        if client:
            client.set_gui_state(bool(int_state))
            
        self.patchbay_manager.optional_gui_state_changed(
            client_id, bool(int_state))

    def _ray_gui_client_still_running(self, path, args):
        client_id = args[0]
        client = self.get_client(client_id)
        if client:
            client.allow_kill()

    def _ray_gui_client_no_save_level(self, path, args):
        client_id, no_save_level = args

        client = self.get_client(client_id)
        if client:
            client.set_no_save_level(no_save_level)

    def _ray_gui_trash_add(self, path, args):
        trashed_client = TrashedClient(self)
        trashed_client.update(*args)
        trash_action = self.main_win.trash_add(trashed_client)
        trashed_client.set_menu_action(trash_action)
        self.trashed_clients.append(trashed_client)

    def _ray_gui_trash_ray_hack_update(self, path, args: list):
        client_id = args.pop(0)
        for trashed_client in self.trashed_clients:
            if trashed_client.client_id == client_id:
                trashed_client.ray_hack = ray.RayHack.new_from(*args)
                break

    def _ray_gui_trash_ray_net_update(self, path, args: list):
        client_id = args.pop(0)
        for trashed_client in self.trashed_clients:
            if trashed_client.client_id == client_id:
                trashed_client.ray_net = ray.RayNet.new_from(*args)
                break

    def _ray_gui_trash_remove(self, path, args):
        client_id = args[0]

        for trashed_client in self.trashed_clients:
            if trashed_client.client_id == client_id:
                break
        else:
            return

        self.trashed_clients.remove(trashed_client)
        self.main_win.trash_remove(trashed_client.menu_action)

    def _ray_gui_trash_clear(self, path, args):
        self.trashed_clients.clear()
        self.main_win.trash_clear()

    def _ray_gui_favorites_added(self, path, args):
        template_name, icon_name, int_factory, display_name = args

        for favorite in self.favorite_list:
            if (favorite.name == template_name
                    and favorite.factory == bool(int_factory)):
                # favorite already exists, update its contents
                favorite.icon = icon_name
                favorite.display_name = display_name
                break
        else:
            self.favorite_list.append(ray.Favorite(
                template_name, icon_name, bool(int_factory), display_name))
            self.signaler.favorite_added.emit(
                template_name, icon_name, bool(int_factory), display_name)
        self.main_win.update_favorites_menu()

    def _ray_gui_favorites_removed(self, path, args):
        template_name, int_factory = args

        for favorite in self.favorite_list:
            if (favorite.name == template_name
                    and favorite.factory == bool(int_factory)):
                break
        else:
            return

        self.favorite_list.remove(favorite)
        self.signaler.favorite_removed.emit(template_name, bool(int_factory))
        self.main_win.update_favorites_menu()

    def _ray_gui_preview_clear(self, path, args):
        self.preview_notes = ''
        self.preview_client_list.clear()
        self.preview_started_clients.clear()
        self.preview_snapshots.clear()
        self.preview_size = -1

    def _ray_gui_preview_notes(self, path, args):
        self.preview_notes = args[0]
    
    def _ray_gui_preview_client_update(self, path, args):
        client = ray.ClientData.new_from(*args)
        for pv_client in self.preview_client_list:
            if pv_client.client_id == client.client_id:
                pv_client.update(*args)
                break
        else:
            self.preview_client_list.append(client)

    def _ray_gui_preview_client_ray_hack_update(self, path, args):
        client_id = args.pop(0)
        for pv_client in self.preview_client_list:
            if pv_client.client_id == client_id:
                pv_client.set_ray_hack(ray.RayHack.new_from(*args))
                break
    
    def _ray_gui_preview_client_ray_net_update(self, path, args):
        client_id = args.pop(0)
        for pv_client in self.preview_client_list:
            if pv_client.client_id == client_id:
                pv_client.set_ray_net(ray.RayNet.new_from(*args))
                break

    def _ray_gui_preview_client_is_started(self, path, args):
        client_id, is_started = args

        for pv_client in self.preview_client_list:
            if pv_client.client_id == client_id:
                if is_started:
                    self.preview_started_clients.add(client_id)
                break

    def _ray_gui_preview_snapshot(self, path, args):
        self.preview_snapshots.append(args[0])

    def _ray_gui_preview_session_size(self, path, args):
        self.preview_size = args[0]

    def _ray_gui_script_info(self, path, args):
        text = args[0]
        self.main_win.show_script_info(text)

    def _ray_gui_hide_script_info(self, path, args):
        self.main_win.hide_script_info_dialog()

    def _ray_gui_script_user_action(self, path, args):
        text = args[0]
        self.main_win.show_script_user_action_dialog(text)

    def _ray_gui_hide_script_user_action(self, path, args):
        self.main_win.hide_script_user_action_dialog()

    def _ray_gui_patchbay_announce(self, path, args):
        self.patchbay_manager.patchbay_announce(*args)

    def _ray_gui_patchbay_client_name_and_uuid(self, path, args):
        self.patchbay_manager.set_group_uuid_from_name(*args)

    def _ray_gui_patchbay_port_added(self, path, args):
        self.patchbay_manager.add_port(*args)

    def _ray_gui_patchbay_port_removed(self, path, args):
        self.patchbay_manager.remove_port(*args)

    def _ray_gui_patchbay_port_renamed(self, path, args):
        self.patchbay_manager.rename_port(*args)
        
    def _ray_gui_patchbay_metadata_updated(self, path, args):
        self.patchbay_manager.metadata_update(*args)

    def _ray_gui_patchbay_connection_added(self, path, args):
        self.patchbay_manager.add_connection(*args)

    def _ray_gui_patchbay_connection_removed(self, path, args):
        self.patchbay_manager.remove_connection(*args)

    def _ray_gui_patchbay_update_group_position(self, path, args):
        self.patchbay_manager.update_group_position(*args)

    def _ray_gui_patchbay_update_portgroup(self, path, args):
        self.patchbay_manager.update_portgroup(*args)

    def _ray_gui_patchbay_server_started(self, path, args):
        self.patchbay_manager.server_started(*args)

    def _ray_gui_patchbay_server_stopped(self, path, args):
        self.patchbay_manager.server_stopped(*args)

    def _ray_gui_patchbay_server_lose(self, path, args):
        self.patchbay_manager.server_lose(*args)

    def _ray_gui_patchbay_dsp_load(self, path, args):
        self.patchbay_manager.set_dsp_load(*args)

    def _ray_gui_patchbay_add_xrun(self, path, args):
        self.patchbay_manager.add_xrun(*args)

    def _ray_gui_patchbay_buffer_size(self, path, args):
        self.patchbay_manager.buffer_size_changed(*args)

    def _ray_gui_patchbay_sample_rate(self, path, args):
        self.patchbay_manager.sample_rate_changed(*args)

    def _ray_gui_patchbay_transport_position(self, path, args):
        self.patchbay_manager.refresh_transport(TransportPosition(
            args[0], bool(args[1]), bool(args[2]), *args[3:]))

    def _ray_gui_patchbay_big_packets(self, path, args):
        self.patchbay_manager.receive_big_packets(*args)

    def _ray_gui_patchbay_fast_temp_file_memory(self, path, args):
        self.patchbay_manager.fast_temp_file_memory(*args)

    def _ray_gui_patchbay_fast_temp_file_running(self, path, args):
        self.patchbay_manager.fast_temp_file_running(*args)



from ctypes import (
    ARRAY,
    c_int,
    c_uint8,
    Structure,
    c_float,
    c_uint32,
    c_uint64,
    c_size_t,
    pointer,
    POINTER,
    c_int32,
    c_double,
    c_char_p,
    CFUNCTYPE,
    c_void_p
)


# Pre-Types
_c_enum = c_int
_c_uchar = c_uint8

class _jack_port(Structure):
    _fields_ = []


class _jack_client(Structure):
    _fields_ = []


# Types
jack_client_t = _jack_client
jack_default_audio_sample_t = c_float
jack_midi_data_t = _c_uchar
jack_nframes_t = c_uint32
jack_port_id_t = c_uint32
jack_port_t = _jack_port
jack_time_t = c_uint64
jack_unique_t = c_uint64
jack_uuid_t = c_uint64

jack_options_t = _c_enum  # JackOptions
jack_status_t = _c_enum  # JackStatus
jack_transport_state_t = _c_enum  # JackTransportState
jack_position_bits_t = _c_enum  # JackPositionBits
jack_session_event_type_t = _c_enum  # JackSessionEventType
jack_session_flags_t = _c_enum  # JackSessionFlags
jack_latency_callback_mode_t = _c_enum  # JackLatencyCallbackMode
jack_property_change_t = _c_enum  # JacKPropertyChange

# JACK2 only:
jack_port_type_id_t = c_uint32

# Structs
class jack_midi_event_t(Structure):
    time: jack_nframes_t
    size: c_size_t
    buffer: 'pointer[jack_midi_data_t]'
    _fields_ = [
        ("time", jack_nframes_t),
        ("size", c_size_t),
        ("buffer", POINTER(jack_midi_data_t))]


class jack_latency_range_t(Structure):
    min: jack_nframes_t
    max: jack_nframes_t
    _fields_ = [("min", jack_nframes_t), ("max", jack_nframes_t)]


class jack_position_t(Structure):
    unique_1: jack_unique_t
    usecs: jack_time_t
    frame_rate: jack_nframes_t
    frame: jack_nframes_t
    valid: jack_position_bits_t
    bar: c_int32
    beat: c_int32
    tick: c_int32
    bar_start_tick: c_double
    beats_per_bar: c_float
    beat_type: c_float
    ticks_per_beat: c_double
    beats_per_minute: c_double
    frame_time: c_double
    next_time: c_double
    bbt_offset: jack_nframes_t
    audio_frames_per_video_frame: c_float
    video_offset: jack_nframes_t
    padding: ARRAY(c_int32, 7)
    unique_2: jack_unique_t
    _fields_ = [
        ("unique_1", jack_unique_t),
        ("usecs", jack_time_t),
        ("frame_rate", jack_nframes_t),
        ("frame", jack_nframes_t),
        ("valid", jack_position_bits_t),
        ("bar", c_int32),
        ("beat", c_int32),
        ("tick", c_int32),
        ("bar_start_tick", c_double),
        ("beats_per_bar", c_float),
        ("beat_type", c_float),
        ("ticks_per_beat", c_double),
        ("beats_per_minute", c_double),
        ("frame_time", c_double),
        ("next_time", c_double),
        ("bbt_offset", jack_nframes_t),
        ("audio_frames_per_video_frame", c_float),
        ("video_offset", jack_nframes_t),
        ("padding", ARRAY(c_int32, 7)),
        ("unique_2", jack_unique_t),
    ]


class jack_session_event_t(Structure):
    _fields_ = [
        ("type", jack_session_event_type_t),
        ("session_dir", c_char_p),
        ("client_uuid", c_char_p),
        ("command_line", c_char_p),
        ("flags", jack_session_flags_t),
        ("future", c_uint32),
    ]


class jack_session_command_t(Structure):
    _fields_ = [
        ("uuid", c_char_p),
        ("client_name", c_char_p),
        ("command", c_char_p),
        ("flags", jack_session_flags_t),
    ]


class jack_property_t(Structure):
    key: c_char_p
    data: c_char_p
    type: c_char_p
    _fields_ = [("key", c_char_p), ("data", c_char_p), ("type", c_char_p)]


class jack_description_t(Structure):
    subject: jack_uuid_t
    property_cnt: c_uint32
    properties: 'pointer[jack_property_t]'
    property_size: c_uint32
    _fields_ = [
        ("subject", jack_uuid_t),
        ("property_cnt", c_uint32),
        ("properties", POINTER(jack_property_t)),
        ("property_size", c_uint32),
    ]


# Callbacks
JackThreadCallback = CFUNCTYPE(c_void_p, c_void_p)
JackTimebaseCallback = CFUNCTYPE(
    None,
    jack_transport_state_t,
    jack_nframes_t,
    POINTER(jack_position_t),
    c_int,
    c_void_p
)
JackPropertyChangeCallback = CFUNCTYPE(
    None, jack_uuid_t, c_char_p, jack_property_change_t, c_void_p
)
JackErrorCallback = CFUNCTYPE(None, c_char_p)


import sys
from ctypes import (
    ARRAY,
    CDLL,
    cdll,
    c_char_p,
    POINTER,
    c_int,
    c_void_p,
    c_float,
    c_ulong,
    c_uint32,
    c_size_t,
    sizeof)

from .types import (
    jack_client_t,
    jack_port_t,
    jack_port_id_t,
    jack_port_type_id_t,
    jack_options_t,
    jack_status_t,
    jack_nframes_t,
    jack_uuid_t,
    jack_time_t,
    jack_latency_callback_mode_t,
    jack_transport_state_t,
    jack_midi_data_t,
    jack_session_event_type_t,
    jack_description_t,
    jack_position_t,
    jack_midi_event_t,
    jack_session_event_t,
    jack_session_command_t,
    jack_latency_range_t,
    JackThreadCallback,
    JackErrorCallback,
    JackTimebaseCallback
)

def _is_jack2(jlib: CDLL) -> bool:
    try:
        if jlib.jack_get_version_string:
            return True

    except AttributeError:
        return False
    
    return False

def _is_python_64bit():
    return sizeof(c_void_p) == 8

def _set_cdll_functions(jlib: CDLL):
    try:
        jlib.jack_get_version_string.argtypes = None
        jlib.jack_get_version_string.restype = c_char_p
    except AttributeError:
        jlib.jack_get_version_string = None

    try:
        jlib.jack_client_open.argtypes = [
            c_char_p, jack_options_t, POINTER(jack_status_t), c_char_p]
        jlib.jack_client_open.restype = POINTER(jack_client_t)
    except AttributeError:
        jlib.jack_client_open = None

    try:
        jlib.jack_client_rename.argtypes = [POINTER(jack_client_t), c_char_p]
        jlib.jack_client_rename.restype = c_char_p
    except AttributeError:
        jlib.jack_client_rename = None

    try:
        jlib.jack_client_close.argtypes = [POINTER(jack_client_t)]
        jlib.jack_client_close.restype = c_int
    except AttributeError:
        jlib.jack_client_close = None

    try:
        jlib.jack_client_name_size.argtypes = None
        jlib.jack_client_name_size.restype = c_int
    except AttributeError:
        jlib.jack_client_name_size = None

    try:
        jlib.jack_get_client_name.argtypes = [POINTER(jack_client_t)]
        jlib.jack_get_client_name.restype = c_char_p
    except AttributeError:
        jlib.jack_get_client_name = None

    try:
        jlib.jack_activate.argtypes = [POINTER(jack_client_t)]
        jlib.jack_activate.restype = c_int
    except AttributeError:
        jlib.jack_activate = None

    try:
        jlib.jack_deactivate.argtypes = [POINTER(jack_client_t)]
        jlib.jack_deactivate.restype = c_int
    except AttributeError:
        jlib.jack_deactivate = None

    try:
        jlib.jack_get_client_pid.argtypes = [c_char_p]
        jlib.jack_get_client_pid.restype = c_int
    except AttributeError:
        jlib.jack_get_client_pid = None

    try:
        jlib.jack_is_realtime.argtypes = [POINTER(jack_client_t)]
        jlib.jack_is_realtime.restype = c_int
    except AttributeError:
        jlib.jack_is_realtime = None
        
def _set_non_callback_cdll_func(jlib: CDLL):
    try:
        jlib.jack_cycle_wait.argtypes = [POINTER(jack_client_t)]
        jlib.jack_cycle_wait.restype = jack_nframes_t
    except AttributeError:
        jlib.jack_cycle_wait = None

    try:
        jlib.jack_cycle_signal.argtypes = [POINTER(jack_client_t), c_int]
        jlib.jack_cycle_signal.restype = None
    except AttributeError:
        jlib.jack_cycle_signal = None

    try:
        jlib.jack_set_process_thread.argtypes = [
            POINTER(jack_client_t), JackThreadCallback, c_void_p]
        jlib.jack_set_process_thread.restype = c_int
    except AttributeError:
        jlib.jack_set_process_thread = None
        
def _set_server_control_funcs(jlib: CDLL):
    jlib.jack_set_freewheel.argtypes = [POINTER(jack_client_t), c_int]
    jlib.jack_set_freewheel.restype = c_int

    jlib.jack_set_buffer_size.argtypes = [POINTER(jack_client_t), jack_nframes_t]
    jlib.jack_set_buffer_size.restype = c_int

    jlib.jack_get_sample_rate.argtypes = [POINTER(jack_client_t)]
    jlib.jack_get_sample_rate.restype = jack_nframes_t

    jlib.jack_get_buffer_size.argtypes = [POINTER(jack_client_t)]
    jlib.jack_get_buffer_size.restype = jack_nframes_t

    jlib.jack_engine_takeover_timebase.argtypes = [POINTER(jack_client_t)]
    jlib.jack_engine_takeover_timebase.restype = c_int

    jlib.jack_cpu_load.argtypes = [POINTER(jack_client_t)]
    jlib.jack_cpu_load.restype = c_float
    
def _set_port_funcs(jlib: CDLL, is_jack2: bool):
    jlib.jack_port_register.argtypes = [
        POINTER(jack_client_t), c_char_p, c_char_p, c_ulong, c_ulong]
    jlib.jack_port_register.restype = POINTER(jack_port_t)

    jlib.jack_port_unregister.argtypes = [
        POINTER(jack_client_t), POINTER(jack_port_t)]
    jlib.jack_port_unregister.restype = c_int

    jlib.jack_port_get_buffer.argtypes = [
        POINTER(jack_port_t), jack_nframes_t]
    jlib.jack_port_get_buffer.restype = c_void_p

    jlib.jack_port_name.argtypes = [POINTER(jack_port_t)]
    jlib.jack_port_name.restype = c_char_p

    jlib.jack_port_short_name.argtypes = [POINTER(jack_port_t)]
    jlib.jack_port_short_name.restype = c_char_p

    jlib.jack_port_flags.argtypes = [POINTER(jack_port_t)]
    jlib.jack_port_flags.restype = c_int

    jlib.jack_port_type.argtypes = [POINTER(jack_port_t)]
    jlib.jack_port_type.restype = c_char_p

    if is_jack2:
        jlib.jack_port_type_id.argtypes = [POINTER(jack_port_t)]
        jlib.jack_port_type_id.restype = jack_port_type_id_t

    jlib.jack_port_is_mine.argtypes = [
        POINTER(jack_client_t), POINTER(jack_port_t)]
    jlib.jack_port_is_mine.restype = c_int

    jlib.jack_port_connected.argtypes = [POINTER(jack_port_t)]
    jlib.jack_port_connected.restype = c_int

    jlib.jack_port_connected_to.argtypes = [POINTER(jack_port_t), c_char_p]
    jlib.jack_port_connected_to.restype = c_int

    jlib.jack_port_get_connections.argtypes = [POINTER(jack_port_t)]
    jlib.jack_port_get_connections.restype = POINTER(c_char_p)

    jlib.jack_port_get_all_connections.argtypes = [
        POINTER(jack_client_t), POINTER(jack_port_t)]
    jlib.jack_port_get_all_connections.restype = POINTER(c_char_p)

    jlib.jack_port_tie.argtypes = [POINTER(jack_port_t), POINTER(jack_port_t)]
    jlib.jack_port_tie.restype = c_int

    jlib.jack_port_untie.argtypes = [POINTER(jack_port_t)]
    jlib.jack_port_untie.restype = c_int

    jlib.jack_port_set_name.argtypes = [POINTER(jack_port_t), c_char_p]
    jlib.jack_port_set_name.restype = c_int

    jlib.jack_port_set_alias.argtypes = [POINTER(jack_port_t), c_char_p]
    jlib.jack_port_set_alias.restype = c_int

    jlib.jack_port_unset_alias.argtypes = [POINTER(jack_port_t), c_char_p]
    jlib.jack_port_unset_alias.restype = c_int

    jlib.jack_port_get_aliases.argtypes = [
        POINTER(jack_port_t), POINTER(ARRAY(c_char_p, 2))]
    jlib.jack_port_get_aliases.restype = c_int

    jlib.jack_port_request_monitor.argtypes = [POINTER(jack_port_t), c_int]
    jlib.jack_port_request_monitor.restype = c_int

    jlib.jack_port_request_monitor_by_name.argtypes = [
        POINTER(jack_client_t), c_char_p, c_int]
    jlib.jack_port_request_monitor_by_name.restype = c_int

    jlib.jack_port_ensure_monitor.argtypes = [POINTER(jack_port_t), c_int]
    jlib.jack_port_ensure_monitor.restype = c_int

    jlib.jack_port_monitoring_input.argtypes = [POINTER(jack_port_t)]
    jlib.jack_port_monitoring_input.restype = c_int

    jlib.jack_connect.argtypes = [POINTER(jack_client_t), c_char_p, c_char_p]
    jlib.jack_connect.restype = c_int

    jlib.jack_disconnect.argtypes = [
        POINTER(jack_client_t), c_char_p, c_char_p]
    jlib.jack_disconnect.restype = c_int

    jlib.jack_port_disconnect.argtypes = [
        POINTER(jack_client_t), POINTER(jack_port_t)]
    jlib.jack_port_disconnect.restype = c_int

    jlib.jack_port_name_size.argtypes = None
    jlib.jack_port_name_size.restype = c_int

    jlib.jack_port_type_size.argtypes = None
    jlib.jack_port_type_size.restype = c_int

    # JACK1 >= 0.125.0, JACK2 >= 1.19.11
    try:
        jlib.jack_port_rename.argtypes = [
            POINTER(jack_client_t), POINTER(jack_port_t), c_char_p]
        jlib.jack_port_rename.restype = c_int
    except AttributeError:
        jlib.jack_port_rename = None

    try:
        jlib.jack_port_type_get_buffer_size.argtypes = [
            POINTER(jack_client_t), c_char_p]
        jlib.jack_port_type_get_buffer_size.restype = c_size_t
    except AttributeError:
        jlib.jack_port_type_get_buffer_size = None

    try:
        jlib.jack_port_uuid.argtypes = [POINTER(jack_port_t)]
        jlib.jack_port_uuid.restype = jack_uuid_t
    except AttributeError:
        jlib.jack_port_uuid = None
        
def _set_latency_func(jlib: CDLL):
    jlib.jack_port_set_latency.argtypes = [POINTER(jack_port_t), jack_nframes_t]
    jlib.jack_port_set_latency.restype = None

    try:
        jlib.jack_port_get_latency_range.argtypes = [
            POINTER(jack_port_t),
            jack_latency_callback_mode_t,
            POINTER(jack_latency_range_t),
        ]
        jlib.jack_port_get_latency_range.restype = None
    except AttributeError:
        jlib.jack_port_get_latency_range = None

    try:
        jlib.jack_port_set_latency_range.argtypes = [
            POINTER(jack_port_t),
            jack_latency_callback_mode_t,
            POINTER(jack_latency_range_t),
        ]
        jlib.jack_port_set_latency_range.restype = None
    except AttributeError:
        jlib.jack_port_set_latency_range = None

    jlib.jack_recompute_total_latencies.argtypes = [POINTER(jack_client_t)]
    jlib.jack_recompute_total_latencies.restype = c_int

    jlib.jack_port_get_latency.argtypes = [POINTER(jack_port_t)]
    jlib.jack_port_get_latency.restype = jack_nframes_t

    jlib.jack_port_get_total_latency.argtypes = [
        POINTER(jack_client_t), POINTER(jack_port_t)]
    jlib.jack_port_get_total_latency.restype = jack_nframes_t

    jlib.jack_recompute_total_latency.argtypes = [
        POINTER(jack_client_t), POINTER(jack_port_t)]
    jlib.jack_recompute_total_latency.restype = c_int
    
def _set_port_searching_func(jlib: CDLL):
    jlib.jack_get_ports.argtypes = [POINTER(jack_client_t), c_char_p, c_char_p, c_ulong]
    jlib.jack_get_ports.restype = POINTER(c_char_p)

    jlib.jack_port_by_name.argtypes = [POINTER(jack_client_t), c_char_p]
    jlib.jack_port_by_name.restype = POINTER(jack_port_t)

    jlib.jack_port_by_id.argtypes = [POINTER(jack_client_t), jack_port_id_t]
    jlib.jack_port_by_id.restype = POINTER(jack_port_t)
    
def _set_time_func(jlib: CDLL):
    jlib.jack_frames_since_cycle_start.argtypes = [POINTER(jack_client_t)]
    jlib.jack_frames_since_cycle_start.restype = jack_nframes_t

    jlib.jack_frame_time.argtypes = [POINTER(jack_client_t)]
    jlib.jack_frame_time.restype = jack_nframes_t

    jlib.jack_last_frame_time.argtypes = [POINTER(jack_client_t)]
    jlib.jack_last_frame_time.restype = jack_nframes_t

    try:
        # JACK_OPTIONAL_WEAK_EXPORT
        jlib.jack_get_cycle_times.argtypes = [
            POINTER(jack_client_t),
            POINTER(jack_nframes_t),
            POINTER(jack_time_t),
            POINTER(jack_time_t),
            POINTER(c_float),
        ]
        jlib.jack_get_cycle_times.restype = c_int
    except AttributeError:
        jlib.jack_get_cycle_times = None

    jlib.jack_frames_to_time.argtypes = [POINTER(jack_client_t), jack_nframes_t]
    jlib.jack_frames_to_time.restype = jack_time_t

    jlib.jack_time_to_frames.argtypes = [POINTER(jack_client_t), jack_time_t]
    jlib.jack_time_to_frames.restype = jack_nframes_t

    jlib.jack_get_time.argtypes = None
    jlib.jack_get_time.restype = jack_time_t
    
def _set_misc_func(jlib: CDLL):
    jlib.jack_free.argtypes = [c_void_p]
    jlib.jack_free.restype = None

    try:
        jlib.jack_set_error_function.argtypes = [JackErrorCallback]
        jlib.jack_set_error_function.restype = None
    except AttributeError:
        jlib.jack_set_error_function = None
        
def _set_transport_func(jlib: CDLL):
    jlib.jack_release_timebase.argtypes = [POINTER(jack_client_t)]
    jlib.jack_release_timebase.restype = c_int

    # jlib.jack_set_sync_callback.argtypes = [POINTER(jack_client_t), JackSyncCallback, c_void_p]
    # jlib.jack_set_sync_callback.restype = c_int

    jlib.jack_set_sync_timeout.argtypes = [
        POINTER(jack_client_t), jack_time_t]
    jlib.jack_set_sync_timeout.restype = c_int

    jlib.jack_set_timebase_callback.argtypes = [
        POINTER(jack_client_t),
        c_int,
        JackTimebaseCallback,
        c_void_p,
    ]
    jlib.jack_set_timebase_callback.restype = c_int

    jlib.jack_transport_locate.argtypes = [
        POINTER(jack_client_t), jack_nframes_t]
    jlib.jack_transport_locate.restype = c_int

    jlib.jack_transport_query.argtypes = [
        POINTER(jack_client_t), POINTER(jack_position_t)]
    jlib.jack_transport_query.restype = jack_transport_state_t

    jlib.jack_get_current_transport_frame.argtypes = [POINTER(jack_client_t)]
    jlib.jack_get_current_transport_frame.restype = jack_nframes_t

    jlib.jack_transport_reposition.argtypes = [
        POINTER(jack_client_t), POINTER(jack_position_t)]
    jlib.jack_transport_reposition.restype = c_int

    jlib.jack_transport_start.argtypes = [POINTER(jack_client_t)]
    jlib.jack_transport_start.restype = None

    jlib.jack_transport_stop.argtypes = [POINTER(jack_client_t)]
    jlib.jack_transport_stop.restype = None
    
def _set_midi_func(jlib: CDLL):
    jlib.jack_midi_get_event_count.argtypes = [c_void_p]
    jlib.jack_midi_get_event_count.restype = jack_nframes_t

    jlib.jack_midi_event_get.argtypes = [POINTER(jack_midi_event_t), c_void_p, c_uint32]
    jlib.jack_midi_event_get.restype = c_int

    jlib.jack_midi_clear_buffer.argtypes = [c_void_p]
    jlib.jack_midi_clear_buffer.restype = None

    jlib.jack_midi_max_event_size.argtypes = [c_void_p]
    jlib.jack_midi_max_event_size.restype = c_size_t

    jlib.jack_midi_event_reserve.argtypes = [c_void_p, jack_nframes_t, c_size_t]
    jlib.jack_midi_event_reserve.restype = POINTER(jack_midi_data_t)

    jlib.jack_midi_event_write.argtypes = [
        c_void_p,
        jack_nframes_t,
        POINTER(jack_midi_data_t),
        c_size_t,
    ]
    jlib.jack_midi_event_write.restype = c_int

    jlib.jack_midi_get_lost_event_count.argtypes = [c_void_p]
    jlib.jack_midi_get_lost_event_count.restype = c_uint32
    
def _set_session_func(jlib: CDLL):    
    try:
        jlib.jack_session_reply.argtypes = [
            POINTER(jack_client_t), POINTER(jack_session_event_t)]
        jlib.jack_session_reply.restype = c_int
    except AttributeError:
        jlib.jack_session_reply = None

    try:
        jlib.jack_session_event_free.argtypes = [
            POINTER(jack_session_event_t)]
        jlib.jack_session_event_free.restype = None
    except AttributeError:
        jlib.jack_session_event_free = None

    try:
        jlib.jack_client_get_uuid.argtypes = [POINTER(jack_client_t)]
        jlib.jack_client_get_uuid.restype = c_char_p
    except AttributeError:
        jlib.jack_client_get_uuid = None

    try:
        jlib.jack_session_notify.argtypes = [
            POINTER(jack_client_t),
            c_char_p,
            jack_session_event_type_t,
            c_char_p,
        ]
        jlib.jack_session_notify.restype = POINTER(jack_session_command_t)
    except AttributeError:
        jlib.jack_session_notify = None

    try:
        jlib.jack_session_commands_free.argtypes = [
            POINTER(jack_session_command_t)]
        jlib.jack_session_commands_free.restype = None
    except AttributeError:
        jlib.jack_session_commands_free = None

    try:
        jlib.jack_get_uuid_for_client_name.argtypes = [
            POINTER(jack_client_t), c_char_p]
        jlib.jack_get_uuid_for_client_name.restype = c_char_p
    except AttributeError:
        jlib.jack_get_uuid_for_client_name = None

    try:
        jlib.jack_get_client_name_by_uuid.argtypes = [
            POINTER(jack_client_t), c_char_p]
        jlib.jack_get_client_name_by_uuid.restype = c_char_p
    except AttributeError:
        jlib.jack_get_client_name_by_uuid = None

    try:
        jlib.jack_reserve_client_name.argtypes = [
            POINTER(jack_client_t), c_char_p, c_char_p]
        jlib.jack_reserve_client_name.restype = c_int
    except AttributeError:
        jlib.jack_reserve_client_name = None

    try:
        jlib.jack_client_has_session_callback.argtypes = [
            POINTER(jack_client_t), c_char_p]
        jlib.jack_client_has_session_callback.restype = c_int
    except AttributeError:
        jlib.jack_client_has_session_callback = None

    try:
        jlib.jack_uuid_parse.argtypes = [c_char_p, POINTER(jack_uuid_t)]
        jlib.jack_uuid_parse.restype = c_int
    except AttributeError:
        jlib.jack_uuid_parse = None

    try:
        jlib.jack_uuid_unparse.argtypes = [jack_uuid_t, c_char_p]
        jlib.jack_uuid_unparse.restype = None
    except AttributeError:
        jlib.jack_uuid_unparse = None
        
def _set_metadata_func(jlib: CDLL):
    try:
        jlib.jack_free_description.argtypes = [
            POINTER(jack_description_t), c_int]
        jlib.jack_free_description.restype = None

        jlib.jack_get_all_properties.argtypes = [
            POINTER(POINTER(jack_description_t))]
        jlib.jack_get_all_properties.restype = c_int

        jlib.jack_get_properties.argtypes = [
            jack_uuid_t, POINTER(jack_description_t)]
        jlib.jack_get_properties.restype = c_int

        jlib.jack_get_property.argtypes = [
            jack_uuid_t, c_char_p, POINTER(c_char_p), POINTER(c_char_p)]
        jlib.jack_get_property.restype = c_int

        jlib.jack_remove_all_properties.argtypes = [POINTER(jack_client_t)]
        jlib.jack_remove_all_properties.restype = c_int

        jlib.jack_remove_properties.argtypess = [
            POINTER(jack_client_t), POINTER(jack_uuid_t)]
        jlib.jack_remove_properties.restype = c_int

        jlib.jack_remove_property.argtypes = [
            POINTER(jack_client_t), POINTER(jack_uuid_t), c_char_p]
        jlib.jack_remove_property.restype = c_int

        jlib.jack_set_property.argtypes = [
            POINTER(jack_client_t),
            jack_uuid_t,
            c_char_p,
            c_char_p,
            c_char_p,
        ]
        jlib.jack_set_property.restype = c_int

    except AttributeError:
        jlib.jack_free_description = None
        jlib.jack_get_properties = None
        jlib.jack_get_property = None
        jlib.jack_remove_all_properties = None
        jlib.jack_remove_properties = None
        jlib.jack_remove_property = None
        jlib.jack_set_property = None
    
def get_jlib() -> CDLL:
    # Load JACK shared library
    try:
        if sys.platform == "darwin":
            _libname = "libjack.dylib"
        elif sys.platform in ("win32", "cygwin"):
            if _is_python_64bit():
                _libname = "libjack64.dll"
            else:
                _libname = "libjack.dll"
        else:
            _libname = "libjack.so.0"

        jlib = cdll.LoadLibrary(_libname)
    except OSError:
        raise ImportError("JACK is not available in this system")
    
    is_jack2 = _is_jack2(jlib)
    _set_cdll_functions(jlib)
    _set_non_callback_cdll_func(jlib)
    _set_server_control_funcs(jlib)
    _set_port_funcs(jlib, is_jack2)
    _set_latency_func(jlib)
    _set_port_searching_func(jlib)
    _set_time_func(jlib)
    _set_misc_func(jlib)
    _set_transport_func(jlib)
    _set_midi_func(jlib)
    _set_session_func(jlib)
    _set_metadata_func(jlib)
    return jlib



from typing import TYPE_CHECKING
from PyQt5.QtCore import QPoint

from . import patchcanvas
from .patchcanvas import CallbackAct, PortMode, PortType, BoxLayoutMode
from .base_elements import Port, GroupPosFlag, PortTypesViewFlag, PortgroupMem
from .port_info_dialog import CanvasPortInfoDialog

if TYPE_CHECKING:
    from .patchbay_manager import PatchbayManager


class Callbacker:
    ''' manage actions coming from patchcanvas.
        Each protected method implements the action to run
        when the action happens. '''
    
    def __init__(self, manager: 'PatchbayManager'):
        self.mng = manager
        self.patchcanvas = patchcanvas
    
    def receive(self, action: CallbackAct, args: tuple):
        ''' receives a callback from patchcanvas and execute
            the function with action name in lowercase.'''
        func_name = '_' + action.name.lower()
        if func_name in self.__dir__():
            self.__getattribute__(func_name)(*args)
    
    # ￬￬￬ functions connected to CallBackAct ￬￬￬
    
    def _group_info(self, group_id: int):
        pass
    
    def _group_rename(self, group_id: int):
        pass
    
    def _group_split(self, group_id: int):        
        group = self.mng.get_group_from_id(group_id)
        if group is not None:
            on_place = not bool(
                group.current_position.flags & GroupPosFlag.HAS_BEEN_SPLITTED)
            self.patchcanvas.split_group(
                group_id, on_place=on_place)
            group.current_position.flags |= GroupPosFlag.SPLITTED
            group.current_position.flags |= GroupPosFlag.HAS_BEEN_SPLITTED
            group.save_current_position()

    def _group_join(self, group_id: int):
        self.patchcanvas.animate_before_join(group_id)
    
    def _group_joined(self, group_id: int):
        group = self.mng.get_group_from_id(group_id)
        if group is not None:
            group.current_position.flags &= ~GroupPosFlag.SPLITTED
            group.save_current_position()
    
    def _group_move(self, group_id: int, port_mode: PortMode, x: int, y: int):
        group = self.mng.get_group_from_id(group_id)
        if group is not None:
            gpos = group.current_position

            if port_mode is PortMode.NULL:
                gpos.null_xy = (x, y)
            elif port_mode is PortMode.INPUT:
                gpos.in_xy = (x, y)
            elif port_mode is PortMode.OUTPUT:
                gpos.out_xy = (x, y)

            group.save_current_position()
    
    def _group_wrap(self, group_id: int, splitted_mode, yesno: bool):
        group = self.mng.get_group_from_id(group_id)
        if group is not None:
            group.wrap_box(splitted_mode, yesno)
    
    def _group_layout_change(self, group_id: int, port_mode: PortMode,
                            layout_mode: BoxLayoutMode):
        group = self.mng.get_group_from_id(group_id)
        if group is not None:
            group.set_layout_mode(port_mode, layout_mode)
    
    def _portgroup_add(self, group_id: int, port_mode: PortMode,
                       port_type: PortType, port_ids: tuple[int]):
        port_list = list[Port]()
        above_metadatas = False

        for port_id in port_ids:
            port = self.mng.get_port_from_id(group_id, port_id)
            if port.mdata_portgroup:
                above_metadatas = True
            port_list.append(port)

        group = self.mng.get_group_from_id(group_id)
        if group is None:
            return

        # we add a PortgroupMem, manager will add the portgroup with it
        pg_mem = PortgroupMem()
        pg_mem.group_name = group.name
        pg_mem.port_type = port_type
        pg_mem.port_mode = port_mode
        
        pg_mem.above_metadatas = above_metadatas
        pg_mem.port_names = [p.short_name() for p in port_list]

        self.mng.add_portgroup_memory(pg_mem)
        self.mng.save_portgroup_memory(pg_mem)
    
    def _portgroup_remove(self, group_id: int, portgroup_id: int):
        group = self.mng.get_group_from_id(group_id)
        if group is None:
            return

        for portgroup in group.portgroups:
            if portgroup.portgroup_id == portgroup_id:
                for port in portgroup.ports:
                    # save a fake portgroup with one port only
                    # it will be considered as a forced mono port
                    # (no stereo detection)
                    pg_mem = PortgroupMem()
                    pg_mem.group_name = group.name
                    pg_mem.port_type = port.type
                    pg_mem.port_mode = portgroup.port_mode
                    pg_mem.above_metadatas = bool(port.mdata_portgroup)
                    pg_mem.port_names = [port.short_name()]
                    self.mng.add_portgroup_memory(pg_mem)
                    self.mng.save_portgroup_memory(pg_mem)
                break

    def _port_info(self, group_id: int, port_id: int):
        port = self.mng.get_port_from_id(group_id, port_id)
        if port is None:
            return

        dialog = CanvasPortInfoDialog(self.mng.main_win)
        dialog.set_port(port)
        dialog.show()

    def _port_rename(self, group_id: int, port_id: int):
        pass
    
    def _ports_connect(self, group_out_id: int, port_out_id: int,
                       group_in_id: int, port_in_id: int):
        pass

    def _ports_disconnect(self, connection_id: int):
        pass

    def _bg_right_click(self, x: int, y: int):
        if self.mng.canvas_menu is not None:
            self.mng.canvas_menu.exec(QPoint(x, y))
    
    def _bg_double_click(self):
        self.mng.sg.full_screen_toggle_wanted.emit()
    
    def _client_show_gui(self, group_id: int, visible: int):
        pass
                    
    def _theme_changed(self, theme_ref: str):
        if self.mng.options_dialog is not None:
            self.mng.options_dialog.set_theme(theme_ref)

        self.mng.remove_and_add_all()